/*******************************************************************
 * 
 *  Author: Peter van Sebille (peter@yipton.net)
 *
 *  (c) Copyright 2016, Peter van Sebille
 *  All Rights Reserved
 *
 *******************************************************************/

#include "mediadirectorymapper.h"
#include "fileutils.h"
#include "stringutils.h"
#include "platform.h"

std::string PartialConvertName(const std::string& aName, char aIgnoreLeading, char aIgnoreTrailing)
{
	if ((aIgnoreLeading == 0) && (aIgnoreTrailing == 0))
	{
		auto startPos = aName.rfind('.');
		if (startPos != std::string::npos)
			return StringUtils::GetPureAlphaNumName(std::string(aName, 0, startPos)); // copy excluding file extension
		return StringUtils::GetPureAlphaNumName(aName);
	}
	
	size_t startPos = 0;
	if (aIgnoreLeading)
	{
		startPos = aName.find(aIgnoreLeading);
		if (startPos != std::string::npos)
		{
			// kill spaces after the match
			startPos++;
			while (aName[startPos] == ' ')
				startPos++;
		}
		else
			startPos = 0;
	}
	
	size_t endPos = std::string::npos;
	if (aIgnoreTrailing)
	{ 
		endPos = aName.find(aIgnoreTrailing, 0);
		if (endPos != std::string::npos)
		{
			// kill spaces before the match
			while (aName[endPos-1] == ' ')
				endPos--;
		}
	}
	return std::string(aName, startPos, endPos - startPos);
}


std::string ConvertName(const std::string& aName, char aIgnoreLeading, char aIgnoreTrailing)
{
	return StringUtils::GetPureAlphaNumName(PartialConvertName(aName, aIgnoreLeading, aIgnoreTrailing));
}

std::string MediaDirectoryMapper::GetMediaMapFileName(const std::string& aDirectory)
{
	std::string result = aDirectory;
	StringUtils::Substitute(result, Platform::PathSeperator(), "_", StringUtils::TSubstituteOptions::EAll);
	StringUtils::Substitute(result, ":", "_", StringUtils::TSubstituteOptions::EAll);
	return result + ".cache";
}

void MediaDirectoryMapper::AddDirectory(const TMediaDirectory& aMediaDirectory, const TGameMap& aGameMap, const std::string& aUserDir)
{
	std::string mediaMapFileName = aUserDir + Platform::PathSeperator() + GetMediaMapFileName(aMediaDirectory.iPathName);
	
	if (FileUtils::FileExists(mediaMapFileName))
		ReadMediaMapFile(aMediaDirectory.iPathName, mediaMapFileName, aGameMap);
	else
		CreateMediaMapFile(aMediaDirectory, mediaMapFileName, aGameMap);
}

void MediaDirectoryMapper::CreateMediaMapFile(const TMediaDirectory& aMediaDirectory, const std::string& aMediaMapFileName, const TGameMap& aGameMap)
{	
	TStringList dirList;
	FileUtils::ReadDirectory(aMediaDirectory.iPathName, dirList);
	
	printf("Creating media map file for directory: %s\n", aMediaDirectory.iPathName.c_str());
	for(auto &dirEntry : dirList)
	{
		std::string convertedDirEntry = ConvertName(dirEntry, aMediaDirectory.GetIgnoreLeadingChar(), aMediaDirectory.GetIngoreTrailingChar());
		if (aGameMap.find(convertedDirEntry) != aGameMap.end())
			iMapper[aMediaDirectory.iPathName][convertedDirEntry] = dirEntry;
//		else
//			printf("  NOT MAPPED: dir=%s, converted=%s, entry=%s\n", aMediaDirectory.iPathName.c_str(), convertedDirEntry.c_str(), dirEntry.c_str());
	}
	
	if (FileUtils::WriteFile(aMediaMapFileName, iMapper[aMediaDirectory.iPathName]))
	{
		printf("  Media map file: %s\n", aMediaMapFileName.c_str());
		printf("  # directory entries: %d, # entries media map: %d\n", (int) dirList.size(), (int) iMapper[aMediaDirectory.iPathName].size());
	}
	else
		printf("Warning: could not create media map file: %s\n", aMediaMapFileName.c_str());
	
#if 1
	int count = 0;
	for (auto &game : aGameMap)
	{
		if (!HasMediaFile(aMediaDirectory.iPathName, game.second.iName))
		{
			printf("  %s (%s) has no media in directory %s\n", game.second.iDescription.c_str(), game.second.iName.c_str(), aMediaDirectory.iPathName.c_str());
			count++;
		}
	}
	printf("  %d out of %d games have no media files: in directory %s\n", count, (int) aGameMap.size(), aMediaDirectory.iPathName.c_str());
#endif
}


void MediaDirectoryMapper::AddDirectory(const TMediaDirectory& aMediaDirectory, const std::string& aFileExtension)
{	
	TStringList dirList;
	FileUtils::ReadDirectory(aMediaDirectory.iPathName, dirList);
	
	std::string fileExtension = aFileExtension;
	std::transform(fileExtension.begin(), fileExtension.end(), fileExtension.begin(), ::tolower);
	
	printf("Creating media map file for directory: %s (%s)\n", aMediaDirectory.iPathName.c_str(), fileExtension.c_str());
	for(auto &dirEntry : dirList)
	{
		auto pos = dirEntry.find_last_of('.');
		if (pos != std::string::npos)
		{
			std::string dirEntryExension = std::string(dirEntry.c_str() + pos + 1);
			std::transform(dirEntryExension.begin(), dirEntryExension.end(), dirEntryExension.begin(), ::tolower);
			if (fileExtension.compare(dirEntryExension) != 0)
				continue;
		}
		
		std::string convertedDirEntry = ConvertName(dirEntry, aMediaDirectory.GetIgnoreLeadingChar(), aMediaDirectory.GetIngoreTrailingChar());
		iMapper[aMediaDirectory.iPathName][convertedDirEntry] = dirEntry;
	}
	printf("  # directory entries: %d, # entries media map: %d\n", (int) dirList.size(), (int) iMapper[aMediaDirectory.iPathName].size());
}


void MediaDirectoryMapper::ReadMediaMapFile(const std::string& aDirectory, const std::string& aMediaMapFileName, const TGameMap& aGameMap)
{
	int count = FileUtils::ReadFile(aMediaMapFileName, iMapper[aDirectory]);
	printf("Read %d entries from media map file: %s\n", count, aMediaMapFileName.c_str());
}

