/*******************************************************************
 * 
 *  Author: Peter van Sebille (peter@yipton.net)
 *
 *  (c) Copyright 2016, Peter van Sebille
 *  All Rights Reserved
 *
 *******************************************************************/

#include "systemmenucontroller.h"
#include "sdlsimpletextwidget.h"
#include "gamemenus.h"
#include "sdlaudio.h"
#include "settingsdialogs.h"
#include "sdlinfoprint.h"
#include "resourcefileuserproperties.h"
#include "resourcefilejoystickproperties.h"
#include "sdluserproperties.h"
#include "appmenudefs.h"
#include "thememodel.h"
#include "vlcwidget.h"


SystemMenuController::SystemMenuController(GameMenus& aGameMenus, sdl::Audio& aAudio)
: iGameMenus(aGameMenus) , iAudio(aAudio)
{
	iHandleMenuItemMethods = 
	{
		{ KMenuCmdConfigureJoystickNavigation, &SystemMenuController::CmdConfigureJoystickNavigation},
		{ KMenuCmdConfigureJoystickScrollSpeed, &SystemMenuController::CmdConfigureJoystickScrollSpeed},
		{ KMenuCmdConfigureJoystickButtons, &SystemMenuController::CmdConfigureJoystickButtons},
		{ KMenuCmdFilterCategories, &SystemMenuController::CmdFilterCategories},
		{ KMenuCmdFilterGameEngines, &SystemMenuController::CmdFilterGameEngines},
		{ KMenuCmdAudioEvents, &SystemMenuController::CmdAudioEvents},
		{ KMenuCmdAudioVolume, &SystemMenuController::CmdAudioVolume},
		{ KMenuCmdVideoSettings, &SystemMenuController::CmdVideoSettings},
		{ KMenuCmdAbout, &SystemMenuController::CmdAbout}
	};
}


void SystemMenuController::SaveResourceFileUserProperties()
{
	if (ResourceFileUserProperties::Instance.Save())
		sdl::InfoPrint::Message(std::string("Settings saved"));
	else
		sdl::InfoPrint::Message(std::string("Error: could not save file: ") + ResourceFileUserProperties::Instance.GetFileName());
}

void SystemMenuController::SaveResourceFileJoystickProperties()
{
	if (ResourceFileJoystickProperties::Instance.Save())
		sdl::InfoPrint::Message(std::string("Settings saved"));
	else
		sdl::InfoPrint::Message("Error: could not save file: " + ResourceFileJoystickProperties::Instance.GetFileName());
}


void SystemMenuController::MenuItemCommand(const sdl::MenuItem& aMenuItem, const sdl::TCommand& aCommand)
{
	if (aCommand == sdl::TCommand::ESelect)
	{
		auto it = iHandleMenuItemMethods.find(aMenuItem.GetName());
		if (it != iHandleMenuItemMethods.end())
			(this->*it->second)();
	}
}


void SystemMenuController::CmdFilterCategories()
{
	DialogFilterCategories dialogFilterCategories(iGameMenus.GetCategoriesList(), AppUserProperties.iGameCategories.iExclude);
	dialogFilterCategories.SetTitle(KMenuCmdFilterCategories);
	if (dialogFilterCategories.Execute())
	{
		SaveResourceFileUserProperties();
		iGameMenus.UpdateCategories();
	}
}

void SystemMenuController::CmdFilterGameEngines()
{
	DialogFilterGameEngines dialog(iGameMenus.GetGameEngineNames(), AppUserProperties.iGameEngines.iExclude);
	dialog.SetTitle(KMenuCmdFilterGameEngines);
	if (dialog.Execute())
	{
		SaveResourceFileUserProperties();
		iGameMenus.RePopulate();
	}
}


void SystemMenuController::CmdConfigureJoystickNavigation()
{
	std::vector<int*> buttonIds = 
	{
		&sdl::SDLJoystickProperties.Navigation.iHatLeft, 
		&sdl::SDLJoystickProperties.Navigation.iHatRight, 
		&sdl::SDLJoystickProperties.Navigation.iHatUp, 
		&sdl::SDLJoystickProperties.Navigation.iHatDown
	};
	std::vector<std::string> buttonTexts = {"Left", "Right", "Up", "Down"};
	
	DialogConfigureJoystickNavigation dialogConfigureJoystickNavigation(buttonTexts, buttonIds, sdl::SDLJoystickProperties.Navigation.iJoystick);
	dialogConfigureJoystickNavigation.SetTitle(KMenuCmdConfigureJoystickNavigation);
	if (dialogConfigureJoystickNavigation.Execute())
		SaveResourceFileJoystickProperties();
}

void SystemMenuController::CmdConfigureJoystickScrollSpeed()
{
	DialogConfigureJoystickNavigationSpeed dialogConfigureJoystickNavigationSpeed(
		sdl::SDLJoystickProperties.Navigation.Speed.iScrollSpeedInMs, sdl::SDLJoystickProperties.Navigation.Speed.iFastScrollSpeedInMs, sdl::SDLJoystickProperties.Navigation.Speed.iFastScrollDelayInMs);
	dialogConfigureJoystickNavigationSpeed.SetTitle(KMenuCmdConfigureJoystickScrollSpeed);
	if (dialogConfigureJoystickNavigationSpeed.Execute())
		SaveResourceFileJoystickProperties();
}

void SystemMenuController::CmdConfigureJoystickButtons()
{
	DialogConfigureJoystickButtons dialogConfigureJoystickButtons(sdl::SDLJoystickProperties.iButtons);
	dialogConfigureJoystickButtons.SetTitle(KMenuCmdConfigureJoystickButtons);
	if (dialogConfigureJoystickButtons.Execute())
		SaveResourceFileJoystickProperties();
}

void SystemMenuController::CmdAbout()
{
#if 0
	TestDialog dialog;
	dialog.SetTitle("TEST");
	dialog.Execute();
	return;
#endif	
	DialogAbout dialogAbout;
	dialogAbout.SetTitle("About");
	dialogAbout.Execute();
}

static void GetDescriptionList(const std::vector<sdl::Audio::TAudioData>& aAudioData, std::vector<std::string>& aDescriptionList)
{
	aDescriptionList.clear();
	for (auto &audioData : aAudioData)
		aDescriptionList.push_back(audioData.iDescription);
};

static void SetEnabledEventDescriptions(
	const std::vector<sdl::Audio::TAudioData>& aAudioData, 
	const std::vector<std::string> aDescriptionList,
	std::vector<std::string>& aEnabledEvents)
{
	aEnabledEvents.clear();
	if ((sdl::SDLUserProperties.iAudioProperties.iEnabledEvents.size() == 1) && (sdl::SDLUserProperties.iAudioProperties.iEnabledEvents.at(0).compare("all") == 0))
		aEnabledEvents = aDescriptionList;
	
	// if not set - convert the audio file name list in sdl::SDLUserProperties.iAudioPropertiesEvents.iEnabledEvents to the audio descriptions
	if (aEnabledEvents.size() == 0)
		for (auto &fileName : sdl::SDLUserProperties.iAudioProperties.iEnabledEvents)
			for (auto &audioData : aAudioData)
				if (audioData.iFileName == fileName)
					aEnabledEvents.push_back(audioData.iDescription);
};

static void SetConfigAudioEvents(
	const std::vector<sdl::Audio::TAudioData>& aAudioData, 
	const std::vector<std::string>& aEnabledEvents)
{
	sdl::SDLUserProperties.iAudioProperties.iEnabledEvents.clear();
	if (aEnabledEvents.size() == aAudioData.size())
		sdl::SDLUserProperties.iAudioProperties.iEnabledEvents.push_back("all");
	else
		for (auto &enabledEvent : aEnabledEvents)
			for (auto &audioData : aAudioData)
				if (audioData.iDescription == enabledEvent)
					sdl::SDLUserProperties.iAudioProperties.iEnabledEvents.push_back(audioData.iFileName);
}

void SystemMenuController::CmdAudioEvents()
{
	std::vector<sdl::Audio::TAudioData> audioData;
	iAudio.GetAudioData(audioData);
	
	std::vector<std::string> descriptionList;
	GetDescriptionList(audioData, descriptionList);
	
	std::vector<std::string> enabledEvents;
	SetEnabledEventDescriptions(audioData, descriptionList, enabledEvents);
	
	DialogAudioEvents dialogAudioEvents(descriptionList, enabledEvents);
	dialogAudioEvents.SetTitle(KMenuCmdAudioEvents);
	if (dialogAudioEvents.Execute())
	{
		SetConfigAudioEvents(audioData, enabledEvents);
		
		SaveResourceFileUserProperties();
	}
}

void SystemMenuController::CmdAudioVolume()
{
	DialogAudioVolume dialogAudioVolume(sdl::SDLUserProperties.iAudioProperties.iVolume, sdl::SDLUserProperties.iAudioProperties.iMute);
	dialogAudioVolume.SetTitle(KMenuCmdAudioVolume);
	if (dialogAudioVolume.Execute())
	{
		SaveResourceFileUserProperties();
		iAudio.SetVolume(sdl::SDLUserProperties.iAudioProperties.iVolume);
	}
}


void SystemMenuController::CmdVideoSettings()
{
	DialogVideoSettings dialogVideoSettings(
		AppUserProperties.iVideoSetting.iVideoEnabled, 
		AppUserProperties.iVideoSetting.iMuteAudio, 
		AppUserProperties.iVideoSetting.iAudioVolumeDuringVideo, 
		AppUserProperties.iVideoSetting.iStartAfterDelay,
		AppUserProperties.iVideoSetting.iTimeLabel
	);
	dialogVideoSettings.SetTitle(KMenuCmdVideoSettings);
	if (dialogVideoSettings.Execute())
	{
		SaveResourceFileUserProperties();
		VlcWidgetAudioControl::SetMute(AppUserProperties.iVideoSetting.iMuteAudio);
		VlcWidgetAudioControl::SetAudioLevel(AppUserProperties.iVideoSetting.iAudioVolumeDuringVideo);
	}
}
