/*******************************************************************
 * 
 *  Author: Peter van Sebille (peter@yipton.net)
 *
 *  (c) Copyright 2016, Peter van Sebille
 *  All Rights Reserved
 *
 *******************************************************************/

#include "fileutils.h"
#include "platform.h"


#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>

class IFileStore
{
public:
	virtual void Store(std::string& aKey, std::string& aValue) = 0;
};

class TStringListFileStore : public IFileStore
{
public:
	TStringListFileStore(TStringList& aStringList) : iStringList(aStringList) {}
	virtual void Store(std::string& aKey, std::string& aValue){iStringList.push_back(aKey);}
	TStringList& iStringList;
};

class TStringMapFileStore : public IFileStore
{
public:
	TStringMapFileStore(TStringMap& aStringMap) : iStringMap(aStringMap) {}
	virtual void Store(std::string& aKey, std::string& aValue){iStringMap[aKey] = aValue;}
	TStringMap& iStringMap;
};


class TStringPairListFileStore : public IFileStore
{
public:
	TStringPairListFileStore(TStringPairList& aStringPairList) : iStringPairList(aStringPairList) {}
	virtual void Store(std::string& aKey, std::string& aValue){iStringPairList.push_back(std::make_pair(aKey, aValue));}
	TStringPairList& iStringPairList;
};


int FileUtils::ReadFile(const std::string& aFileName, TStringList& aStringList, char aSeperator)
{
	TStringListFileStore stringListFileStore(aStringList);
	return ReadFile(aFileName, stringListFileStore, aSeperator);
}

int FileUtils::ReadFile(const std::string& aFileName, TStringMap& aStringMap, char aSeperator)
{
	TStringMapFileStore stringMapFileStore(aStringMap);
	return ReadFile(aFileName, stringMapFileStore, aSeperator);
}

int FileUtils::ReadFile(const std::string& aFileName, TStringPairList& aStringPairList, char aSeperator)
{
	TStringPairListFileStore stringPairListFileStore(aStringPairList);

	return ReadFile(aFileName, stringPairListFileStore, aSeperator);
}


int FileUtils::ReadDirectory(const std::string& aDirName, TStringList& aStringList)
{
	return ReadDirectory(aDirName, aStringList, [=](struct dirent* aDirEnt, std::string& aOutName){ aOutName = aDirEnt->d_name; return true;});
}

static bool DoExtensionsMatch(struct dirent* aDirEntry, const std::string& aFileExtension)
{
	char* p = strstr(aDirEntry->d_name, aFileExtension.c_str());
	return (p && (p != aDirEntry->d_name) && (p[-1]=='.'));
}

int FileUtils::ReadDirectory(const std::string& aDirName, TStringList& aStringList, const std::string& aFileExtension)
{
	return ReadDirectory(aDirName, aStringList, [aFileExtension](struct dirent* aDirEnt, std::string& aOutName){ if (::DoExtensionsMatch(aDirEnt, aFileExtension)) {aOutName = aDirEnt->d_name; return true;} return false;});
}

int FileUtils::ReadDirectory(const std::string& aDirName, TStringList& aStringList, const std::function <bool (struct dirent*, std::string& aOutName)>& aWantEntryFunction)
{
	DIR*			dir;
	struct dirent*	ent;
	
	if ((dir = opendir (aDirName.c_str())) != nullptr) 
	{
		while ((ent = readdir (dir)) != nullptr) 
		{
			std::string name;
			if (strcmp(ent->d_name, ".") == 0)
				continue;
			if (strcmp(ent->d_name, "..") == 0)
				continue;
			if (aWantEntryFunction(ent, name))
				aStringList.push_back(name);
		}
		closedir (dir);
	}
	
	return aStringList.size();
}


int FileUtils::ReadFile(const std::string& aFileName, IFileStore& aFileStore, char aSeperator)
{
	std::ifstream file(aFileName.c_str(), std::ios::in);
	if (!file.good())
	{
		std::cout << "Warning: could not open " << aFileName << "\n";
		return 0;
	}
	
	std::string line;
	int count = 0;
	while (std::getline(file, line))
	{
		count++;

		if (line.size() == 0)
				continue;

		std::istringstream stream(line);
		std::string key;
		std::string value;

		if(std::getline(stream, key, aSeperator))
			std::getline(stream, value);
		else
			key.assign(line);

		aFileStore.Store(key, value);
	}

	return count;
}

bool FileUtils::FileExists(const std::string& aFileName)
{
	return Platform::FileExists(aFileName);
}


bool FileUtils::DirectoryExists(const std::string& aDirName)
{
	return Platform::DirectoryExists(aDirName);
}

bool FileUtils::WriteFile(const std::string& aFileName, const TStringPairList& aStringPairList)
{
	std::ofstream file(aFileName.c_str(), std::ios::out | std::ios::binary | std::ios::trunc);
	if (!file.good())
	{
		std::cout << "ERROR: could not open file " << aFileName << " for writing. Bailing out.\n";
		return false;
	}
	
	for(auto &stringPair : aStringPairList)
		file << stringPair.first << "=" << stringPair.second << std::endl;

	file.close();
	
	return true;
}

bool FileUtils::WriteFile(const std::string& aFileName, const TStringMap& aStringMap)
{
	std::ofstream file(aFileName.c_str(), std::ios::out | std::ios::binary | std::ios::trunc);
	if (!file.good())
	{
		std::cout << "ERROR: could not open file " << aFileName << " for writing. Bailing out.\n";
		return false;
	}
	
	for(auto &stringPair : aStringMap)
		file << stringPair.first << "=" << stringPair.second << std::endl;

	file.close();
	
	return true;
}


void FileUtils::ReadFilePart(const std::string& aFileName, int aOffset, int aLength, std::string& aContent)
{
	aContent.clear();
	
	char* buf = (char*) malloc(aLength + 1);
	if (buf == nullptr)
		return;
	
	std::unique_ptr<char, void (*)(void*)> bufPtr(buf, free);
	
	std::ifstream file(aFileName.c_str(), std::ios::in);
	if (!file.good())
	{
		std::cout << "ERROR: could not open " << aFileName << ". Bailing out.\n";
		return;
	}

	file.seekg(aOffset, std::ios::beg);
	file.read(buf, aLength);
	buf[aLength] = 0;
	aContent.assign(buf);
}

void FileUtils::FindFile(const std::string& aFileName, const std::string& aDirectoryName, std::string& aFoundFileName)
{
	// First open aFileName in aDirectoryName
	aFoundFileName = aDirectoryName + Platform::PathSeperator() + aFileName; 
	if (!FileUtils::FileExists(aFoundFileName))
	{
		// Alternatively open aFileName as is
		aFoundFileName = aFileName;
		if (!FileUtils::FileExists(aFoundFileName))
			aFoundFileName.clear();
	}
}

bool FileUtils::FindFileInDirs(const std::string& aFileName, const std::vector<std::string> aDirs, std::string& aFoundFileName)
{
	// first check filename as is - then search for it in the aDirs directories
	aFoundFileName = aFileName;
	bool found = FileUtils::FileExists(aFoundFileName);
	if (!found)
	{
		for (auto &dir : aDirs)
		{
			aFoundFileName = dir + Platform::PathSeperator() + aFileName;
			if (FileUtils::FileExists(aFoundFileName))
			{
				found = true;
				break;
			}
		}
	}
	if (!found)
		aFoundFileName.clear();

	return found;
}
