/*******************************************************************
 * 
 *  Author: Peter van Sebille (peter@yipton.net)
 *
 *  (c) Copyright 2016, Peter van Sebille
 *  All Rights Reserved
 *
 *******************************************************************/

#ifndef __RESOURCES_H
#define __RESOURCES_H

#include "platform.h"
#include "types.h"

template <class T>
void Encode(std::string& aEncodedData, T& aRawResourceItem);

template <class T>
bool Decode(const std::string& aEncodedData, T& aRawResourceItem);

template<class T>
void Defaulter(T&);

class Resources
{
public:
	void Reset() { iResources.clear(); }
	template <class T>
	void Add(const std::string& aKey, T& aValue, void aFunction(T&) = nullptr, int aRound = 0) 
	{
		iResources[aKey] = std::unique_ptr<ResourceItemBase>(new ResourceItem<T>(aValue, aFunction, aRound)); 
	}
	
	bool Decode(const std::string& aKey, const std::string& aValue, bool& aKeyNotFound);
	void Encode(std::vector<std::string>& aResourceNames, std::vector<std::string>& aEncodedStrings) const;
	void ApplyDefaults()
	{
		for (auto &resource  : iResources)
			resource.second.get()->ApplyDefault(0);
		for (auto &resource  : iResources)
			resource.second.get()->ApplyDefault(1);
	}

private:
	class ResourceItemBase
	{
	public:
		virtual ~ResourceItemBase() { }
		virtual void Encode(std::string& aEncodedData) = 0;
		virtual bool Decode(const std::string& aEncodedData) = 0;
		virtual void ApplyDefault(int aRound) = 0;
	};

	template <class T>
	class ResourceItem : public ResourceItemBase
	{
	public:
		ResourceItem(T& aRawResourceItem, void aFunction(T&) = nullptr, int aRound = 0) 
			: iRawResourceItem(aRawResourceItem), iFunction(aFunction), iRound(aRound) { }

		virtual void Encode(std::string& aEncodedData) override
		{
			::Encode(aEncodedData, iRawResourceItem);
		}
		
		virtual bool Decode(const std::string& aEncodedData) override
		{
			return ::Decode(aEncodedData, iRawResourceItem);
		}
		
		virtual void ApplyDefault(int aRound)
		{
			if (iFunction && (aRound == iRound))
				iFunction(iRawResourceItem);
		}
		
		T& iRawResourceItem;
		void (*iFunction)(T& aValue) = nullptr;
		int iRound = 0;
	};
	
	typedef std::map<std::string, std::unique_ptr<ResourceItemBase>> TResources;
	TResources iResources;
};


#endif // __RESOURCES_H