/*******************************************************************
 * 
 *  Author: Peter van Sebille (peter@yipton.net)
 *
 *  (c) Copyright 2016, Peter van Sebille
 *  All Rights Reserved
 *
 *******************************************************************/

#include "stringutils.h"

bool StringUtils::Substitute(std::string& aText, const std::string& aMatchString, const std::string& aSubstituteString, StringUtils::TSubstituteOptions aSubstituteOptions)
{
	bool substituted = false;
	
	size_t pos = aText.find(aMatchString);
	while ((pos !=  std::string::npos) && (((aSubstituteOptions==StringUtils::TSubstituteOptions::EFirst) && !substituted) || (aSubstituteOptions==StringUtils::TSubstituteOptions::EAll)))
	{
		substituted = true;
		aText.replace(pos, aMatchString.size(), aSubstituteString);
		pos = aText.find(aMatchString);
	}
	
	return substituted;
}

void StringUtils::Substitute(std::string& aText, const TSubstituteMap& aSubstituteMap)
{
	for (auto &mapEntry : aSubstituteMap)
		while (Substitute(aText, mapEntry.first, mapEntry.second))
			;
}

void StringUtils::SplitString(const std::string& aString, std::vector<std::string>& aStringList)
{
	aStringList.clear();
	std::stringstream ss(aString);
	std::string string;
	while (std::getline(ss, string, ',')) 
		aStringList.push_back(string);
}


static bool hasJustWhiteSpaces(const std::string& aString, int aSize)
{
	for (int i=0 ; i<aSize; i++)
	{
		char c = aString.at(i);
		if (!(c==' ' || c=='\t'))
			return false;
	}
	return true;
}

bool StringUtils::StripQuotes(std::string& aString)
{
	bool success = true;
	int startQuotePos = aString.find_first_of('\"');
	if ((startQuotePos != (int) std::string::npos) && hasJustWhiteSpaces(aString, startQuotePos-1))
	{
		int endQuotePos = aString.find_last_of('\"');
		if (endQuotePos > startQuotePos)
			aString = std::string(aString, startQuotePos+1, endQuotePos - startQuotePos - 1);
		else
			success = false;
	}
	return success;
}

bool StringUtils::IsSame(const std::string& a, const std::string& b)
{
	unsigned int sz = a.size();
	if (b.size() != sz)
		return false;
	for (unsigned int i = 0; i < sz; ++i)
		if (tolower(a[i]) != tolower(b[i]))
			return false;
	return true;
}

bool StringUtils::GetBoolValueFromString(const std::string& aValue, bool& aBool)
{
	bool success = true;
	
	if (IsSame(aValue, std::string("true")) || IsSame(aValue, std::string("yes")))
		aBool = true;
	else if ((IsSame(aValue, std::string("false")) || IsSame(aValue, std::string("no"))))
		aBool = false;
	else
		success = false;
	
	return success;
}

struct NonAlphaNumChar
{
	bool operator()(char c)
	{
		return !(
			((c>=48) && (c<=57))
			||
			((c>=65) && (c<=90))
			||
			((c>=97) && (c<=122))
		);
	}
};

std::string  StringUtils::GetPureAlphaNumName(const std::string& aString)
{
	std::string pureAlphaNumName = aString;
	pureAlphaNumName.erase(std::remove_if(pureAlphaNumName.begin(), pureAlphaNumName.end(), NonAlphaNumChar()), pureAlphaNumName.end());
	std::transform(pureAlphaNumName.begin(), pureAlphaNumName.end(), pureAlphaNumName.begin(), ::tolower);
	
	return pureAlphaNumName;
}


struct RemoveWhiteSpaceChars
{
	bool operator()(char c)
	{
		return (c =='\r' || c =='\t' || c == '\n');
	}
};

void StringUtils::StripWhiteSpaces(std::string& aString)
{
	aString.erase( std::remove_if(aString.begin(), aString.end(), RemoveWhiteSpaceChars()), aString.end());
	
	// removing leading spaces
	size_t start = aString.find_first_not_of(' ');
	if (start != std::string::npos)
		aString.erase(aString.begin(), aString.begin() + start);
	
	// remove trailing spaces
	start = aString.find_last_not_of(' ');
	if (start != std::string::npos)
		aString.erase(aString.begin() + start + 1, aString.end());
}


static void doFitText(StringUtils::ITextMetrics& aTextMetrics, const std::string& aText, std::string& aReturnText, int aMinTextLen, int aMaxTextLen, int aMaxWidth, int& aWidth, int& aHeight, bool aEclipse)
{
	// binary search - so determine middle of min-max
	int diff = (aMaxTextLen - aMinTextLen) / 2;

	// make sure to subtract diff from max, rather than adding to min
	// otherwise for diff==0 we call TTF_SizeUTF8 with the minimum string length (which we already know fits)
	// rather than with the maximum string length (for which we do not know yet whehther it fits)
	int textLen = aMaxTextLen - diff;

	int width, height;
	aReturnText.assign(aText.c_str(), textLen);
	if (aEclipse)
		aReturnText += "...";
	aTextMetrics.GetTextSize(aReturnText.c_str(), width, height);

	if (width == aMaxWidth)
	{
		// BINGO, precise fit :-);
		aWidth = width;
		aHeight = height;
	}
	else if (diff == 0) // final search step
	{
		if (width < aMaxWidth)
		{
			// the maximum string fitted
			// nothing else to do - we return aTextLen, aWidth, aHeight as just calculated
			aWidth = width;
			aHeight = height;
		}
		else
		{
			// the maximum string didn't fit - return minimum string (only eclipse text when there is text)
			aReturnText.assign(aText.c_str(), aMinTextLen);
			if (aEclipse && (aMinTextLen>0))
				aReturnText += "...";
		}
	}
	else if (width < aMaxWidth)
	{
		// try to fit more text, make current aTextLen the minimum string length
		aWidth = width;
		aHeight = height;
		doFitText(aTextMetrics, aText, aReturnText, textLen, aMaxTextLen, aMaxWidth, aWidth, aHeight, aEclipse);
	}
	else
	{
		// text didn't fit - make current aTextLen the maximum string length
		doFitText(aTextMetrics, aText, aReturnText, aMinTextLen, textLen, aMaxWidth, aWidth, aHeight, aEclipse);
	}
}


void StringUtils::FitText(ITextMetrics& aTextMetrics, const std::string& aText, std::string& aReturnText, int aMaxWidth, int& aWidth, int& aHeight, bool aEclipse)
{
	aTextMetrics.GetTextSize(aText.c_str(), aWidth, aHeight);
	if (aWidth < aMaxWidth)
		aReturnText = aText;
	else
		doFitText(aTextMetrics, aText, aReturnText, 0, aText.size(), aMaxWidth, aWidth, aHeight, aEclipse);
}

void StringUtils::ToLower(std::string& aString)
{
	std::transform(aString.begin(), aString.end(), aString.begin(), ::tolower);
}


static void addWord(const char* aWord, int aWordSize, std::vector<std::string>& aStringList)
{
	if ((aWord!= nullptr) && (aWordSize>0))
		aStringList.push_back(std::string(aWord, aWordSize));
}

void StringUtils::ParseCommandLine(const std::string& aString, std::vector<std::string>& aStringList)
{
	aStringList.clear();
	if (aString.size() <= 0)
		return;
	
	bool isInQuotedString = false;
	const char* pc = aString.c_str();
	const char* wordStart = pc;
	
	int wordSize = 0;
	bool done = false;
	bool wordBreak = false;
	do
	{
		switch (*pc)
		{
			case 0:
			{
				addWord(wordStart, wordSize, aStringList);
				done = true;
				break;
			}
				
			case ' ':
			{
				if (isInQuotedString)
				{
					pc++;
					wordSize++;
				}
				else if (!wordBreak)
				{
					addWord(wordStart, wordSize, aStringList);
					pc++;
					wordSize = 0;
					wordStart = nullptr;
					wordBreak = true;
				}
				else
				{
					pc++;
				}
				break;
			}
					
			case '"':
			{
				if (isInQuotedString)
				{
					addWord(wordStart, wordSize, aStringList);
					isInQuotedString = false;
					pc++;
					wordSize = 0;
					wordStart = nullptr;
				}
				else
				{
					wordStart = ++pc;
					wordSize = 0;
					isInQuotedString = true;
					wordBreak = false;
				}
				break;
			}
			
			default:
			{
				if (wordBreak)
				{
					wordStart = pc++;
					wordSize = 1;
					wordBreak = false;
				}
				else
				{
					pc++;
					wordSize++;
				}
				break;
			}
		}
	} while (!done);
}
