/*******************************************************************
 * 
 *  Author: Peter van Sebille (peter@yipton.net)
 *
 *  (c) Copyright 2016, Peter van Sebille
 *  All Rights Reserved
 *
 *******************************************************************/

#include "sdlgpurenderer.h"
#include "commonsdlrenderer.h"

namespace sdl
{
	class SDLGPUTexture : public ITexture
	{
	public:
		virtual ~SDLGPUTexture() override;
		virtual TRect GetRect() override;
		virtual void CopyBytes(void* aBytes, int aSize) override;
		virtual void SetColor(const TColor& aColor) override;
		
		SDLGPUTexture(GPU_Image* aTexture = nullptr) : iTexture(aTexture) {}
		SDLGPUTexture(const TSize& aSize, GPU_FormatEnum aTextureFormat = GPU_FORMAT_RGBA);
		GPU_Image* iTexture = nullptr;
	};
	
	class SDLGPURenderer : public IRenderer
	{
	public:
		SDLGPURenderer(const TSize& aSize, const std::string& aWindowTitle, bool aFullscreen);
		virtual ~SDLGPURenderer() override;
		
		virtual ITexture* CreateTextureFromSVG(const std::string& aSVGText) override;
		virtual ITexture* CreateTextureFromTTF(TTF_Font& aFont, const std::string& aText, const TColor& aColor) override;
		virtual ITexture* CreateTextureFromFile(const std::string& aFileName) override;
		
		virtual ITexture*	CreateMemoryTexture(const TSize& aSize, const char* aChromaFormat = nullptr) override;
		virtual ITexture*	CreateRenderingTexture(const TSize& aSize) override;
		
		virtual IRenderer* CreateRenderer(ITexture& aTexture) override;
		
		virtual void RenderTexture(ITexture& aTexture, const TPoint& aPoint) override;
		virtual void RenderTextureRotated(ITexture& aTexture, const TPoint& aPoint, float aDegrees=0) override;
		virtual void RenderTextureScaled(ITexture& aTexture, const TPoint& aPoint, float aScaleX, float aScaleY) override;
		virtual void RenderFillRect(const TRect& aRect, const TColor& aColor) override;
		virtual void RenderRect(const TRect& aRect, const TColor& aColor) override;
		
		virtual TRect SetClipRect(const TRect& aRect) override;
		
	protected:
		SDLGPURenderer() {}
		GPU_Image* CreateTextureFromSurface(SDL_Surface* aSurface);
		virtual void Clear() override;
		virtual void Present() override;
		
		GPU_Target* iRenderer = nullptr;
		
	private:
		void WorkAroundClippingBug();
	};
}


using namespace sdl;

class SDLGPUTextureRenderer : public SDLGPURenderer
{
public:
	SDLGPUTextureRenderer(SDLGPUTexture& aTexture)
	{
		iRenderer = GPU_LoadTarget(aTexture.iTexture);
	}
	
	~SDLGPUTextureRenderer()
	{
		if (iRenderer)
		{
			GPU_FreeTarget(iRenderer);
			iRenderer = nullptr; // must set to nullptr, otherwise the baseclass will call GPU_Quit
		}
	}
};

SDLGPUTexture::SDLGPUTexture(const TSize& aSize, GPU_FormatEnum aTextureFormat)
{
	iTexture = GPU_CreateImage(aSize.iWidth, aSize.iHeight, aTextureFormat);
}

SDLGPUTexture::~SDLGPUTexture() 
{
	if (iTexture)
	{
		GPU_FreeImage(iTexture);
		iTexture = nullptr;
	}
}

TRect SDLGPUTexture::GetRect() 
{
	if (iTexture)
		return TRect(TPoint(0, 0), TSize(iTexture->w, iTexture->h));
	return TRect(0, 0, 0, 0);
}

void SDLGPUTexture::CopyBytes(void* aBytes, int aBytesPerRow) 
{
	GPU_UpdateImageBytes(iTexture, nullptr, (unsigned char*) aBytes, aBytesPerRow); 			
}

void SDLGPUTexture::SetColor(const TColor& aColor)
{
	GPU_SetRGBA(iTexture, aColor.GetR(), aColor.GetG(), aColor.GetB(), aColor.GetA());
}

SDLGPURenderer::SDLGPURenderer(const TSize& aSize, const std::string& aWindowTitle, bool aFullscreen)
{
	GPU_SetDebugLevel(GPU_DEBUG_LEVEL_MAX);
	
	Uint32 sdlWindowFlags = SDL_WINDOW_SHOWN;
	if (aFullscreen)
		sdlWindowFlags = SDL_WINDOW_FULLSCREEN_DESKTOP;
	
	iRenderer = GPU_Init(aSize.iWidth, aSize.iHeight, sdlWindowFlags);
	
	if(iRenderer == nullptr)
	{
		printf("GPU_Init failed\n");
		exit(-1);
	}
	
	WorkAroundClippingBug();
	
	GPU_SetDefaultAnchor(0, 0);
}

SDLGPURenderer::~SDLGPURenderer() 
{
	if (iRenderer)
	{
		GPU_Quit();
		iRenderer = nullptr;
	}
}

IRenderer* SDLGPURenderer::CreateRenderer(ITexture& aTexture)
{
	SDLGPUTexture& texture = (SDLGPUTexture&) aTexture;
	return new SDLGPUTextureRenderer(texture);
}

ITexture* SDLGPURenderer::CreateTextureFromSVG(const std::string& aSVGText) 
{
	void* surfaceData = nullptr;
	SDL_Surface* surface = CreateSurfaceFromSVG(aSVGText, &surfaceData);
	ITexture* texture = new SDLGPUTexture(CreateTextureFromSurface(surface));
	free(surfaceData);
	return texture;
}

ITexture* SDLGPURenderer::CreateTextureFromTTF(TTF_Font& aFont, const std::string& aText, const TColor& aColor) 
{
	SDL_Surface* surface = CreateSurfaceFromTTF(aFont, aText, aColor);
	return new SDLGPUTexture(CreateTextureFromSurface(surface));
}

ITexture* SDLGPURenderer::CreateTextureFromFile(const std::string& aFileName) 
{
	SDL_Surface* surface = CreateSurfaceFromFile(aFileName);
	return new SDLGPUTexture(CreateTextureFromSurface(surface));
}

static std::map<std::string, GPU_FormatEnum> sChromaFormatMapping = {{"RGBA", GPU_FORMAT_RGBA}};

static GPU_FormatEnum GetTextureFormat(const char* aChromaFormat)
{
	GPU_FormatEnum format = GPU_FORMAT_RGBA;
	if (aChromaFormat)
	{
		auto it = sChromaFormatMapping.find(aChromaFormat);
		if (it != sChromaFormatMapping.end())
			format = it->second;
	}
	return format;
}

ITexture* SDLGPURenderer::CreateMemoryTexture(const TSize& aSize, const char* aChromaFormat)
{
	return new SDLGPUTexture(aSize, GetTextureFormat(aChromaFormat));
}

ITexture* SDLGPURenderer::CreateRenderingTexture(const TSize& aSize)
{
	return new SDLGPUTexture(aSize);
}

void SDLGPURenderer::RenderTexture(ITexture& aTexture, const TPoint& aPoint) 
{
	RenderTextureRotated(aTexture, aPoint);
}

void SDLGPURenderer::RenderTextureRotated(ITexture& aTexture, const TPoint& aPoint, float aDegrees) 
{
	SDLGPUTexture& texture = (SDLGPUTexture&) aTexture;
	
	float anchor_x, anchor_y;
	GPU_GetAnchor(texture.iTexture, &anchor_x, &anchor_y);
	GPU_SetAnchor(texture.iTexture, 0.5, 0.5);
	
	TSize size = aTexture.GetRect().Size();
	TPoint p(aPoint.iX + size.iWidth/2, aPoint.iY + size.iHeight/2);
	
	GPU_BlitRotate(texture.iTexture, nullptr, iRenderer, p.iX, p.iY, aDegrees);
	
	GPU_SetAnchor(texture.iTexture, anchor_x, anchor_y);
	
}

void SDLGPURenderer::RenderTextureScaled(ITexture& aTexture, const TPoint& aPoint, float aScaleX, float aScaleY)
{
	SDLGPUTexture& texture = (SDLGPUTexture&) aTexture;
	GPU_BlitScale(texture.iTexture, nullptr, iRenderer, aPoint.iX, aPoint.iY, aScaleX, aScaleX);
}

void SDLGPURenderer::RenderFillRect(const TRect& aRect, const TColor& aColor) 
{
	GPU_RectangleFilled(iRenderer, aRect.Position().iX, aRect.Position().iY, aRect.GetBottomRight().iX, aRect.GetBottomRight().iY, aColor);
}

void SDLGPURenderer::RenderRect(const TRect& aRect, const TColor& aColor) 
{
	GPU_Rectangle(iRenderer, aRect.Position().iX, aRect.Position().iY, aRect.GetBottomRight().iX, aRect.GetBottomRight().iY, aColor);
}

TRect SDLGPURenderer::SetClipRect(const TRect& aRect) 
{
	GPU_Rect rect = aRect;
	rect = GPU_SetClipRect(iRenderer, rect);
	return TRect(rect);
}

GPU_Image* SDLGPURenderer::CreateTextureFromSurface(SDL_Surface* aSurface)
{
	GPU_Image* texture = nullptr;
	if (aSurface)
	{
		texture = GPU_CopyImageFromSurface(aSurface);
		if (texture == nullptr)
			printf("SDLGPURenderer::CreateTextureFromSurface GPU_CopyImageFromSurface failed\n");
		
		SDL_FreeSurface(aSurface);
	}
	
	return texture;
}

void SDLGPURenderer::Clear() 
{
	GPU_Clear(iRenderer);
}

void SDLGPURenderer::Present() 
{
	GPU_Flip(iRenderer);
}

		
void SDLGPURenderer::WorkAroundClippingBug()
{
	iRenderer->clip_rect.w = iRenderer->w;
	iRenderer->clip_rect.h = iRenderer->h;
}


IRenderer* sdl::CreateSDLGPURenderer(const TSize& aSize, const std::string& aWindowTitle, bool aFullScreen)
{
	return new SDLGPURenderer(aSize, aWindowTitle, aFullScreen);
}