/*******************************************************************
 * 
 *  Author: Peter van Sebille (peter@yipton.net)
 *
 *  (c) Copyright 2016, Peter van Sebille
 *  All Rights Reserved
 *
 *******************************************************************/

#include "sdlcore.h"

#include "platform.h"
#include "fileutils.h"
#include "stringutils.h"
#include <nanosvg.h>
#include <nanosvgrast.h>
#include "sdlthemeproperties.h"
#include "sdluserproperties.h"
#include "sdlsystemproperties.h"

using namespace sdl;

/*********************************************
 * 
 * TextMetrics
 *
 *********************************************/


class TextMetrics : public StringUtils::ITextMetrics
{
public:
	TextMetrics(TTF_Font* aFont) : iFont(aFont) {}
	virtual void GetTextSize(const std::string& aText, int& aWidth, int& aHeight) override
	{
		TTF_SizeUTF8(iFont, aText.c_str(), &aWidth, &aHeight);
	}
	TTF_Font* iFont;
};


/*********************************************
 * 
 * FontCache
 * 
 *********************************************/

class FontCache
{
public:
	virtual ~FontCache() { Clear(); }
	TTF_Font* OpenFont(const TFont& aFont);
private:
	void Clear();
	TTF_Font* OpenAndCacheFont(const TFont& aFont);
	TTF_Font* FindCachedFont(const TFont& aFont);
	void PrepareFont(TTF_Font& aTTFFont, const TFont& aFont);
	
	typedef std::map<std::string, TTF_Font*> TFontCache;
	TFontCache iFontCache;
};


/*********************************************
 * 
 * Globals
 * 
 *********************************************/

namespace sdl
{
	ThemeProperties		SDLThemeProperties;
	UserProperties 		SDLUserProperties;
	SystemProperties	SDLSystemProperties;
};


/*********************************************
 * 
 * statics
 * 
 *********************************************/

const std::vector<std::string> KInitialDirs = {"."};

static bool						sReRenderingRequested = false;
static FontCache				sFontCache;

sdl::SystemDirs sdl::System::FontDirs;
sdl::SystemDirs sdl::System::ImageDirs;
sdl::SystemDirs sdl::System::AudioDirs;


/*********************************************
 * 
 * FontCache
 * 
 *********************************************/

void FontCache::Clear()
{
	for (auto &font : iFontCache)
		TTF_CloseFont(font.second);

	iFontCache.clear();
}

TTF_Font* FontCache::OpenFont(const TFont& aFont)
{
	TTF_Font* font = FindCachedFont(aFont);
	if (font == nullptr)
		font = OpenAndCacheFont(aFont);
	if (font)
		PrepareFont(*font, aFont);
	return font;
}

static int GetTextStyle(const TFont& aFont)
{
	int style = TTF_STYLE_NORMAL;
	if (aFont.iBold)
		style |= TTF_STYLE_BOLD;
	if (aFont.iUnderline)
		style |= TTF_STYLE_UNDERLINE;
	if (aFont.iItalic)
		style |= TTF_STYLE_ITALIC;
	if (aFont.iStrikeThrough)
		style |= TTF_STYLE_STRIKETHROUGH;
	
	return style;
}

void FontCache::PrepareFont(TTF_Font& aTTFFont, const TFont& aFont)
{
	TTF_SetFontStyle(&aTTFFont, GetTextStyle(aFont));
	TTF_SetFontHinting(&aTTFFont, TTF_HINTING_NORMAL);

	if ((aFont.iOutline > 0) && (TTF_GetFontOutline(&aTTFFont) != aFont.iOutline))
		TTF_SetFontOutline(&aTTFFont, aFont.iOutline);
}


TTF_Font* FontCache::OpenAndCacheFont(const TFont& aFont)
{
	TTF_Font* font = nullptr;
	std::string fontName;
	
	if (FileUtils::FindFileInDirs(aFont.iName, System::FontDirs.GetDirectoryNames(), fontName))
	{
		font = TTF_OpenFont(fontName.c_str(), aFont.iSize);
		if (font != nullptr)
		{
			std::string fontName(aFont.iName);
			fontName += std::to_string(aFont.iSize);
			
			iFontCache[fontName] = font;
		}
		else
		{
			printf("FontCache::OpenAndCacheFont, could not open font file %s (height=%d): %s\n", fontName.c_str(), aFont.iSize, TTF_GetError());
		}
	}
	else
	{
		printf("FontCache::OpenAndCacheFont, could not find font file %s\n", aFont.iName.c_str());
	}
	
	return font;
}

TTF_Font* FontCache::FindCachedFont(const TFont& aFont)
{
	TTF_Font* font = nullptr;
	
	std::string fontName(aFont.iName);
	fontName += std::to_string(aFont.iSize);
	
	auto entry = iFontCache.find(fontName);
	if(entry != iFontCache.end()) 
		font = entry->second;

	return font;
}



/*********************************************
 * 
 * System
 * 
 *********************************************/


static void DumpVideoInfo()
{
	printf("SDL, video driver list:\n");
	for (int i=0; i<SDL_GetNumVideoDrivers() ; i++)
		printf("  driver %d: %s\n", i, SDL_GetVideoDriver(i));
	printf("SDL: using video driver=%s\n", SDL_GetCurrentVideoDriver());
}


void System::Init()
{
	if (SDL_Init(SDL_INIT_EVERYTHING ) == -1)
	{
		printf("SDL_Init failed: %s\n", SDL_GetError());
		exit(-1);
	}
	
	if (TTF_Init()==-1)
	{
		printf("TTF_Init failed: %s\n", TTF_GetError());
		exit(-1);
	}
	
	SDL_ShowCursor(SDL_DISABLE);
	
	DumpVideoInfo();
}

void System::Shutdown()
{
	SDLThemeProperties.SetDefaults();
	SDLUserProperties.SetDefaults();
	
	System::ImageDirs.Reset();
	System::FontDirs.Reset();
	System::AudioDirs.Reset();
	
	SDL_Quit();
}

void System::RequestReRender()
{
	sReRenderingRequested = true;
}

bool System::IsReRenderRequested()
{
	bool reRenderingRequested = sReRenderingRequested;
	sReRenderingRequested = false;
	return reRenderingRequested;
}


/*********************************************
 * 
 * Drawing svg shapes
 * 
 * I used http://www.drawsvg.org/ to draw basic 
 * SVG shapes and copy & paste the contents here
 * 
 *********************************************/

static const char* svgRoundedRectText  = 
"<svg width=\"%d\" height=\"%d\">"
"<rect x=\"%d\" y=\"%d\" style=\"stroke-width: %d;\" stroke=\"%s\" width=\"%d\" height=\"%d\" fill=\"%s\" fill-opacity=\"%s\" rx=\"%d\" ry=\"%d\"/>"
"</svg>";

static void CreateSVGRoundedRectText(char* aText, const TSize& aSize, const TRoundedRect& aRoundedRect)
{
	char colorFrame[32];
	char colorFill[32];
	
	int delta = (aRoundedRect.iFrameWidth+1)/2;
	
	sprintf(colorFrame, "#%06X", aRoundedRect.iFrameColor.iColor & 0xffffff);
	sprintf(colorFill, "#%06X", aRoundedRect.iFillColor.iColor & 0xffffff);
	bool opacity = aRoundedRect.iFillColor.iColor == -1;
	
	sprintf(aText, svgRoundedRectText, (int) aSize.iWidth, (int) aSize.iHeight, delta, delta,  aRoundedRect.iFrameWidth, colorFrame, 
			(int) aSize.iWidth - aRoundedRect.iFrameWidth, (int) aSize.iHeight - aRoundedRect.iFrameWidth, colorFill, opacity ? "0" : "1", aRoundedRect.iRadiusX, aRoundedRect.iRadiusY);
}

static const char* svgDuoColorFilledRoundedRectText  = 
"<svg width=\"%d\" height=\"%d\">"
"<defs>"
	"<linearGradient gradientUnits=\"objectBoundingBox\" id=\"foo\" spreadMethod=\"pad\" x1=\"0%\" x2=\"100%\" y1=\"0%\" y2=\"0%\">"
	"<stop offset=\"%0.2f\" style=\"stop-color:%s;stop-opacity:1\"/>"
	"<stop offset=\"%0.2f\" style=\"stop-color:%s;stop-opacity:1\"/>"
	"</linearGradient>"
"</defs>"
"<rect x=\"%d\" y=\"%d\" style=\"stroke-width: %d;\" stroke=\"%s\" width=\"%d\" height=\"%d\" fill=\"url(#foo)\" rx=\"%d\" ry=\"%d\"/>"
"</svg>";

static void CreateSVGDuoColorFilledRoundedRectText(char* aText, const TSize& aSize, const TRoundedRect& aRoundedRect, const TColor& aFillColor, int aPercentage)
{
	char colorFrame[32];
	char colorFill1[32];
	char colorFill2[32];
	
	int delta = (aRoundedRect.iFrameWidth+1)/2;
	
	sprintf(colorFrame, "#%06X", aRoundedRect.iFrameColor.iColor & 0xffffff);
	sprintf(colorFill1, "#%06X", aRoundedRect.iFillColor.iColor & 0xffffff);
	sprintf(colorFill2, "#%06X", aFillColor.iColor & 0xffffff);
	
	sprintf(aText, svgDuoColorFilledRoundedRectText, (int) aSize.iWidth, (int) aSize.iHeight, ((float)aPercentage/100), colorFill1,  ((float)aPercentage/100), colorFill2,
			delta, delta,  aRoundedRect.iFrameWidth, colorFrame, 
		 (int) aSize.iWidth - aRoundedRect.iFrameWidth, (int) aSize.iHeight - aRoundedRect.iFrameWidth, aRoundedRect.iRadiusX, aRoundedRect.iRadiusY);
}


static const char* svgArrowRightText  = 
"<svg width=\"%d\" height=\"%d\" viewBox=\"0 0 60 100\">"
"<path d=\"M34.34299850463867,125.65699768066406c3.1240005493164062,3.123992919921875,8.18899917602539,3.123992919921875,11.31399917602539,0l56,-56c1.447998046875,-1.447998046875,2.3430023193359375,-3.447998046875,2.3430023193359375,-5.657001495361328s-0.8949966430664062,-4.2089996337890625,-2.3430023193359375,-5.657001495361328l-56,-56c-3.1240005493164062,-3.124000072479248,-8.18899917602539,-3.124000072479248,-11.31399917602539,0s-3.125,8.189000129699707,0,11.314000129699707l50.34299850463867,50.34299850463867l-50.34299850463867,50.342994689941406c-3.125,3.1240005493164062,-3.125,8.189002990722656,0,11.314002990722656Z\" style=\"vector-effect: non-scaling-stroke; stroke-width: 3px;\" stroke=\"%s\" transform=\"matrix(0.757812, 0, 0, 0.757812, -24.25, 1)\" fill=\"%s\"/>"
"</svg>";

static void CreateSVGArrowRightText(char* aText, int aHeight, const TColor& aFrameColor, const TColor& aFillColor)
{
	char colorFrame[32];
	char colorFill[32];
	
	sprintf(colorFrame, "#%06X", aFrameColor.iColor & 0xffffff);
	sprintf(colorFill, "#%06X", aFillColor.iColor & 0xffffff);
	
	sprintf(aText, svgArrowRightText, (aHeight * 6 ) /10, aHeight, colorFrame, colorFill);
}

static const char* svgCircle  = 
"<svg width=\"%d\" height=\"%d\">"
"<ellipse rx=\"%d\" ry=\"%d\" cx=\"%d\" cy=\"%d\" stroke-width=\"%d\" stroke=\"%s\" fill=\"%s\"/>"
"</svg>";

static void CreateSVGCircleText(char* aText, const TSize& aSize, int aFrameSize, const TColor& aFrameColor, const TColor& aFillColor)
{
	char colorFrame[32];
	char colorFill[32];
	
	sprintf(colorFrame, "#%06X", aFrameColor.iColor & 0xffffff);
	sprintf(colorFill, "#%06X", aFillColor.iColor & 0xffffff);
	
	sprintf(aText, svgCircle, (int) aSize.iWidth, (int) aSize.iHeight, ((int) (aSize.iWidth/2)) - aFrameSize, ((int) (aSize.iHeight/2)) - aFrameSize, (int) aSize.iWidth/2, (int) aSize.iHeight/2, aFrameSize, colorFrame, colorFill);
}



/*********************************************
 * 
 * RenderClipper
 * 
 *********************************************/

RenderClipper::RenderClipper(IRenderer& aRenderer, const TRect& aRect) : iRenderer(aRenderer)
{
	Clip(aRect);
}

RenderClipper::~RenderClipper()
{
	RestoreClip();
}

void RenderClipper::Clip(const TRect& aScreenRect)
{
	iRect = iRenderer.SetClipRect(aScreenRect);
}

void RenderClipper::RestoreClip()
{
	iRenderer.SetClipRect(iRect);
}


/*********************************************
 * 
 * RenderClipper
 * 
 *********************************************/

ITexture* RenderUtils::CreateSVGRoundedRectTexture(IRenderer& aRenderer, const TSize& aSize, const TRoundedRect& aRoundedRect)
{
	char text[4096];
	CreateSVGRoundedRectText(text, aSize, aRoundedRect);
	return aRenderer.CreateTextureFromSVG(text);
}

ITexture* RenderUtils::CreateSVGDuoColorFilledRoundedRectTexture(IRenderer& aRenderer, const TSize& aSize, const TRoundedRect& aRoundedRect, const TColor& aFillColor, int aPercentage)
{
	char text[4096];
	CreateSVGDuoColorFilledRoundedRectText(text, aSize, aRoundedRect, aFillColor, aPercentage);
	return aRenderer.CreateTextureFromSVG(text);
}

ITexture* RenderUtils::CreateSVGArrowRightTexture(IRenderer& aRenderer, int aHeight, const TColor& aFrameColor, const TColor& aFillColor)
{
	char text[4096];
	CreateSVGArrowRightText(text, aHeight, aFrameColor, aFillColor);
	return aRenderer.CreateTextureFromSVG(text);
}

ITexture* RenderUtils::CreateSVGCircleTexture(IRenderer& aRenderer, const TSize& aSize, int aFrameSize, const TColor& aFrameColor, const TColor& aFillColor)
{
	char text[4096];
	CreateSVGCircleText(text, aSize, aFrameSize, aFrameColor, aFillColor);
	return aRenderer.CreateTextureFromSVG(text);
}

ITexture* RenderUtils::CreateTextureFromText(IRenderer& aRenderer, const TFont& aFont, const std::string& aText, int aMaxWidth, bool aEclipseText)
{
	if (aText.size() == 0)
		return nullptr;
	
	ITexture* texture = nullptr;
	
	if (aMaxWidth > 0)
	{
		std::string text = aText;
		int width, height;
		RenderUtils::FitText(aFont, aText, text, aMaxWidth, width, height, aEclipseText);
	
		TTF_Font* ttfFont = sFontCache.OpenFont(aFont);
		if (ttfFont)
			texture = aRenderer.CreateTextureFromTTF(*ttfFont, text.c_str(), aFont.iTextColor);
	}
	
	return texture;
}

void RenderUtils::RenderScaledTextureMaintainAspectRatio(IRenderer& aRenderer, ITexture& aTexture, const TRect& aBoundingRect, TRect& aOutRect)
{
	aOutRect = aTexture.GetRect();
	
	if (aOutRect.Size().IsEmpty())
		aOutRect.SetSize(1);
	
	float diffX = aBoundingRect.Width() / aOutRect.Width();
	float diffY = aBoundingRect.Height() / aOutRect.Height();
	
	float diff = (diffX < diffY) ? diffX : diffY;
	aOutRect = TRect(TSize(aOutRect.Width() * diff, aOutRect.Height() * diff)); 
	aOutRect.Center(aBoundingRect);
	
	aRenderer.RenderTextureScaled(aTexture, aOutRect.Position(), diff, diff);
}

ITexture* RenderUtils::CreateTextureFromFile(IRenderer& aRenderer, const std::string& aFileName)
{ 
	ITexture* texture = nullptr;
	
	std::string imageName;
	
	if (FileUtils::FindFileInDirs(aFileName, System::ImageDirs.GetDirectoryNames(), imageName))
	{
		texture = aRenderer.CreateTextureFromFile(imageName);
	}
	else
		printf("Draw::LoadTexture: could not find image=%s\n", aFileName.c_str());
	
	return texture;
}

void RenderUtils::RenderRect(IRenderer& aRenderer, const TRect& aRect, const TColor& aColor, int aFrameSize)
{
	TRect rect = aRect;
	while (aFrameSize-->0)
	{
		aRenderer.RenderRect(rect, aColor);
		rect.Grow(-1);
	}
}

TSize RenderUtils::RenderText(IRenderer& aRenderer, const TPoint& aPoint, const TFont& aFont, const std::string& aText)
{
	TSize size;
	if (aText.size() == 0)
		return size;
	
	TTF_Font* ttfFont = sFontCache.OpenFont(aFont);
	if (ttfFont)
	{
		ITexture* texture = aRenderer.CreateTextureFromTTF(*ttfFont, aText, aFont.iTextColor);
		if (texture)
		{
			size = texture->GetRect().Size();
			aRenderer.RenderTexture(*texture, aPoint - TPoint(0, size.iHeight));
			delete texture;
			texture = nullptr;
		}
	}
	return size;
}

TSize RenderUtils::GetTextSize(const TFont& aFont, const std::string& aText)
{
	int width = 0;
	int height = 0;
	
	TTF_Font* ttfFont = sFontCache.OpenFont(aFont);
	if (ttfFont)
	{
		if (TTF_SizeUTF8(ttfFont, aText.c_str(), &width, &height) != 0)
			printf("Draw::TextSize() TTF_SizeUTF8 failed for text=\"%s\": %s\n", aText.c_str(), TTF_GetError());
	}
	else
		printf("Draw::TextSize() OpenFont failed for text=\"%s\" %s\n", aText.c_str(), TTF_GetError());
	
	return TSize(width, height);
}

void RenderUtils::FitText(const TFont& aFont, const std::string& aText, std::string& aReturnedText, int aMaxWidth, int& aWidth, int& aHeight, bool aEclipseText)
{
	aWidth = 0;
	aHeight = 0;
	aReturnedText = "";
	
	TTF_Font* ttfFont = sFontCache.OpenFont(aFont);
	if (ttfFont)
	{
		TextMetrics textMetrics(ttfFont);
		StringUtils::FitText(textMetrics, aText, aReturnedText, aMaxWidth, aWidth, aHeight, aEclipseText);
	}
}
	
int RenderUtils::GetDescent(const TFont& aFont)
{
	int descent = 0;
	
	TTF_Font* ttfFont = sFontCache.OpenFont(aFont);
	if (ttfFont)
		descent = TTF_FontDescent(ttfFont);
	
	return descent;
}

	