/*******************************************************************
 * 
 *  Author: Peter van Sebille (peter@yipton.net)
 *
 *  (c) Copyright 2016, Peter van Sebille
 *  All Rights Reserved
 *
 *******************************************************************/

#include "sdldialog.h"
#include "sdluserfeedbackmanager.h"
#include "sdlthemeproperties.h"
#include "sdlsystemrenderer.h"

using namespace sdl;

Dialog::Dialog()
{
	SetDefaultProperties();
	SetDefaultDialogSize();
}

void Dialog::SetDefaultProperties()
{
	iDialogProperties = SDLThemeProperties.iDefaultDialogProperties;
	ApplyProperties();
}

void Dialog::SetDefaultDialogSize()
{
	int width = iDialogProperties.iWidth;
	if (width<=0)
		//width = SDLThemeProperties.iScreen.iWidth / 2;
		width = SDLThemeProperties.iDefaultFont.iSize * 35;
	
	int height = iDialogProperties.iHeight;
	if (height<=0)
		//height = SDLThemeProperties.iScreen.iHeight / 2;
		height = SDLThemeProperties.iDefaultFont.iSize * 20;
	
	SetSize(width, height);
}


bool Dialog::Execute()
{
	// we can't use cached textures to create the faded backgrounded image because we can't wait to be called from Render(). 
	// We're going to ask the event manager to draw the top level widget (to a bitmap). We need to do that before calling
	// EventManager::EventLoop because it will make us the top level widget.
	CreateFadedBackgroundBackupBitmap();
	
	Setup();
	SwitchFocusNext();

	bool success = iEventManager.EventLoop(*this);
	OnDialogExit(success);
	
	return success;
}



void Dialog::CreateFadedBackgroundBackupBitmap()
{
	IRenderer* renderer = SystemRenderer::Instance.GetRenderer();
	if (renderer)
	{
		// create off-screen texture
		ITexture* texture = renderer->CreateRenderingTexture(TSize(SDLThemeProperties.iScreen.iWidth, SDLThemeProperties.iScreen.iHeight));
		
		if (texture)
		{
			IRenderer* offlineRenderer = renderer->CreateRenderer(*texture);
			
			if (offlineRenderer)
			{
				offlineRenderer->Clear();
				
				SystemRenderer::Instance.RenderWidgets(*offlineRenderer);
				
				delete offlineRenderer;
			}
			
			texture->SetColor(iDialogProperties.iBackgroundMask);
			
			iBackupTexture.reset(texture);	
		}
	}
	else
	{
		printf("Dialog::CreateFadedBackgroundBackupBitmap, no renderer\n");
	}
}

void Dialog::Render(IRenderer& aRenderer)
{
	if (iBackupTexture.get())
		aRenderer.RenderTexture(*iBackupTexture.get(), TPoint(0, 0));
	
	Window::Render(aRenderer);
}

bool Dialog::HandleCommandMyself(TCommand aCommand)
{
	bool consumed = true;
	
	switch (aCommand)
	{
		case TCommand::EBack:
		case TCommand::ECancel:
		{
			iEventManager.MarkEventLoopDone(false);
			UserFeedbackManager::SendUserFeedbackEvent(UserFeedbackManager::TEvent::EDialogCancel);
			break;
		}
		
		case TCommand::EOK:
		{
			iEventManager.MarkEventLoopDone(true);
			UserFeedbackManager::SendUserFeedbackEvent(UserFeedbackManager::TEvent::EDialogOk);
			break;
		}
		
		default:
		{
			consumed = Window::HandleCommandMyself(aCommand);
			break;
		}
	}
	return consumed;
}



/*********************************************
 * 
 * SettingsDialog
 * 
 *********************************************/

void SettingsDialog::Layout()
{
	const int spacing = 10;
	const TRect clientRect = TRect(GetClientAreaSize());
	
	int height = iButtonBar.GetMinimumHeight();
	TRect buttonBarRect(clientRect.X(), clientRect.Height() - height, clientRect.Width(), height);
	iButtonBar.SetRect(buttonBarRect);
	iButtonBar.Layout();
	
	TRect listRect = clientRect;
	listRect.SetHeight(listRect.Height() - (height + spacing));
	int numberOfListWidgets = iList.GetNumberOfWidgets();
	if (numberOfListWidgets>0)
		listRect.SetHeight((listRect.Height() / numberOfListWidgets) * numberOfListWidgets);
	iList.SetRect(listRect);
	iList.Layout();
}



/*********************************************
 * 
 * StdDialogs
 * 
 *********************************************/


bool StdDialogs::MessageDialog(const std::string& aTitle, const std::string& aMessage)
{
	SettingsDialog	dialog;
	dialog.SetTitle(aTitle);
	
	LabelWidget label;
	label.SetText(aMessage);
	label.OverrideDefaults(SDLThemeProperties.iDefaultWindowProperties.iLabelTitle);
	
	ButtonWidget button1(TCommand::EOK);
	button1.SetText("OK");
	
	ButtonWidget button2(TCommand::ECancel);
	button2.SetText("Cancel");
	
	dialog.AddButton(button1);
	dialog.AddButton(button2);
	
	dialog.AddInput(label);
	
	dialog.Layout();
	
	return dialog.Execute();
}

