/*******************************************************************
 * 
 *  Author: Peter van Sebille (peter@yipton.net)
 *
 *  (c) Copyright 2016, Peter van Sebille
 *  All Rights Reserved
 *
 *******************************************************************/

#include "sdleventmanager.h"
#include "sdlwidget.h"
#include "sdlsystemrenderer.h"

using namespace sdl;

Widget*	EventManager::iTopLevelWidget = nullptr;
bool	EventManager::iIsLastCommandRepeated = false;

IJoystickHatEventHandler* 		EventManager::iCurrentJoystickHatEventHandler = nullptr;
IJoystickHatEventHandler* 		EventManager::iOldJoystickHatEventHandler = nullptr;
IJoystickButtonEventHandler* 	EventManager::iCurrentJoystickButtonEventHandler = nullptr;
IJoystickButtonEventHandler* 	EventManager::iOldJoystickButtonEventHandler = nullptr;


std::map< std::pair<int, int>, int> EventManager::iKeyTranslateTable = 
{ 
	{{SDLK_RETURN, MAKE_KEY_STATE(false, false)}, (int) TCommand::ESelect}, // keyup, no repeat 
	{{SDLK_ESCAPE, MAKE_KEY_STATE(false, false)}, (int) TCommand::EBack},	// keyup, no repeat 
	
	{{SDLK_UP, MAKE_KEY_STATE(true, true)}, (int) TCommand::EUp},			// keydown, key repeat 
	{{SDLK_DOWN, MAKE_KEY_STATE(true, true)}, (int) TCommand::EDown},		// keydown, key repeat 
	{{SDLK_UP, MAKE_KEY_STATE(true, false)}, (int) TCommand::EUp},			// keydown, no key repeat 
	{{SDLK_DOWN, MAKE_KEY_STATE(true, false)}, (int) TCommand::EDown},		// keydown, no key repeat 
	
	{{SDLK_PAGEUP, MAKE_KEY_STATE(true, false)}, (int) TCommand::EPageUp},			// keydown, no key repeat 
	{{SDLK_PAGEDOWN, MAKE_KEY_STATE(true, false)}, (int) TCommand::EPageDown},		// keydown, no key repeat 
	
	{{SDLK_LEFT, MAKE_KEY_STATE(true, false)}, (int) TCommand::ELeft},		// keydown, no repeat 
	{{SDLK_RIGHT, MAKE_KEY_STATE(true, false)}, (int) TCommand::ERight},	// keydown, no repeat 
	{{SDLK_LEFT, MAKE_KEY_STATE(true, true)}, (int) TCommand::ELeft},		// keydown, repeat 
	{{SDLK_RIGHT, MAKE_KEY_STATE(true, true)}, (int) TCommand::ERight},		// keydown, repeat 
	
	{{SDLK_DELETE, MAKE_KEY_STATE(true, false)}, (int) TCommand::EDelete}	// keydown, no repeat 
};

void EventManager::InjectQuitEvent()
{
	SDL_Event event;
	event.type = SDL_QUIT;
	event.user.code = 0;
	event.user.data1 = event.user.data2 = nullptr;
	SDL_PushEvent(&event);
}

void EventManager::InjectRedrawEvent()
{
	if (!IsSDLEventAlreadyInQueue(EventManager::ESDLEventRedraw))
	{
		SDL_Event event;
		event.type = EventManager::ESDLEventRedraw;
		event.user.code = 0;
		event.user.data1 = event.user.data2 = nullptr;
		SDL_PushEvent(&event);
	}
}

bool EventManager::IsSDLEventAlreadyInQueue(Uint32 aEventID)
{
	SDL_Event event;
	return (SDL_PeepEvents(&event, 1, SDL_PEEKEVENT, aEventID, aEventID) == 1);
}

void EventManager::InjectCustomCommand(TCommand aCommand)
{
	if (!IsSDLEventAlreadyInQueue(EventManager::ESDLEventCommand))
	{
		SDL_Event event;
		event.type = EventManager::ESDLEventCommand;
		event.user.code = (int) aCommand;
		event.user.data1 = event.user.data2 = nullptr;
		SDL_PushEvent(&event);
	}
}

void EventManager::SendCommand(TCommand aCommand)
{
	if (EventManager::iTopLevelWidget != nullptr)
		EventManager::iTopLevelWidget->HandleCommand(aCommand);
}

bool EventManager::IsLastCommandRepeated()
{
	return EventManager::iIsLastCommandRepeated;
}


bool EventManager::EventLoop(Widget& aWidget)
{
	Widget* oldTopLevelWidget = EventManager::iTopLevelWidget;
	EventManager::iTopLevelWidget = &aWidget;
	
	SystemRenderer::Instance.AddWidget(aWidget);
	
	// force a re-render
	System::RequestReRender();
	SystemRenderer::Instance.Render();
	
	iDone = false;
	SDL_Event event;
	while((!iDone && SDL_WaitEvent(&event))) 
	{  
		ProcessEvent(event, aWidget);
		if (System::IsReRenderRequested())
			InjectRedrawEvent();
	}

	SystemRenderer::Instance.RemoveWidget(aWidget);
	
	EventManager::iTopLevelWidget = oldTopLevelWidget;
	
	return iLoopSuccess;
}


std::string GetEventName(const SDL_Event& aEvent)
{
	switch (aEvent.type)
	{
		case SDL_QUIT: return "QUIT";
		case SDL_KEYDOWN: return std::string("KEYDOWN ") + std::to_string(aEvent.key.keysym.sym);
		case SDL_KEYUP: return std::string("KEYUP ") + std::to_string(aEvent.key.keysym.sym);
		case SDL_USEREVENT: return "REDRAW";
		case SDL_JOYHATMOTION: return std::string("JOYHAT, which=") + std::to_string(aEvent.jhat.which) + std::string(" value=") + std::to_string(aEvent.jhat.value);
		case SDL_JOYAXISMOTION: return std::string("JOYAXIS, axis=") + std::to_string(aEvent.jaxis.axis) + std::string(" value=") + std::to_string(aEvent.jaxis.value);
	}
	return std::string("UNKNOWN ") + std::to_string(aEvent.type);
}

static void ConvertJoyAxisEvent(const SDL_JoyAxisEvent& aJoyAxisEvent, SDL_JoyHatEvent& aJoyHatEvent)
{
	aJoyHatEvent.which = aJoyAxisEvent.which;
	aJoyHatEvent.value = SDL_HAT_CENTERED;
	
	if (aJoyAxisEvent.value > 8000)
		aJoyHatEvent.value |= (aJoyAxisEvent.axis == 0)? SDL_HAT_RIGHT : SDL_HAT_UP;
	else if (aJoyAxisEvent.value < -8000)
		aJoyHatEvent.value = (aJoyAxisEvent.axis == 0)? SDL_HAT_LEFT : SDL_HAT_DOWN;
}


void EventManager::ProcessEvent(const SDL_Event& aEvent, Widget& aWidget)
{
	TCommand command = TCommand::ENone;
	
	EventManager::iIsLastCommandRepeated = false;

//	printf("EVENT: %s\n", GetEventName(aEvent).c_str());
	
	if (aEvent.type == SDL_QUIT)
		MarkEventLoopDone();
	else if (aEvent.type == EventManager::ESDLEventRedraw)
		SystemRenderer::Instance.Render();
	else if (aEvent.type == SDL_KEYDOWN)
	{
		auto it = iKeyTranslateTable.find(std::make_pair(aEvent.key.keysym.sym, MAKE_KEY_STATE(true, aEvent.key.repeat != 0)));
		if (it != iKeyTranslateTable.end())
		{
			command = (TCommand) it->second;
			EventManager::iIsLastCommandRepeated = aEvent.key.repeat != 0;
		}
	}
	else if (aEvent.type == SDL_KEYUP)
	{
		auto it = iKeyTranslateTable.find(std::make_pair(aEvent.key.keysym.sym, MAKE_KEY_STATE(false, aEvent.key.repeat != 0)));
		if (it != iKeyTranslateTable.end())
		{
			command = (TCommand) it->second;
			EventManager::iIsLastCommandRepeated = aEvent.key.repeat != 0;
		}
		SDL_FlushEvents(SDL_KEYDOWN, SDL_KEYDOWN);
		SDL_FlushEvents(SDL_KEYUP, SDL_KEYUP);
	}
	else if (aEvent.type == SDL_JOYHATMOTION)
	{
		if (iCurrentJoystickHatEventHandler)
			iCurrentJoystickHatEventHandler->ProcessJoyHatEvent(aEvent.jhat);
	}
	else if (aEvent.type == SDL_JOYAXISMOTION)
	{
		if (iCurrentJoystickHatEventHandler)
		{
			SDL_JoyHatEvent joyHatEvent;
			ConvertJoyAxisEvent(aEvent.jaxis, joyHatEvent);
			
			iCurrentJoystickHatEventHandler->ProcessJoyHatEvent(joyHatEvent);
		}
	}
	else if ((aEvent.type == SDL_JOYBUTTONDOWN) || (aEvent.type == SDL_JOYBUTTONUP))
	{
		if (iCurrentJoystickButtonEventHandler)
			iCurrentJoystickButtonEventHandler->ProcessJoyButtonEvent(aEvent.jbutton);
	}
	else if (aEvent.type == EventManager::ESDLEventCommand)
		command = (TCommand) aEvent.user.code;
	
	if (command != TCommand::ENone)
		aWidget.HandleCommand(command);
}

void EventManager::ClearEvents()
{
	SDL_PumpEvents();
	SDL_FlushEvent(SDL_KEYDOWN);
	SDL_FlushEvent(SDL_KEYUP);
	SDL_FlushEvent(EventManager::ESDLEventCommand);
	SDL_FlushEvent(SDL_JOYHATMOTION);
	SDL_FlushEvent(SDL_JOYAXISMOTION);
	SDL_FlushEvent(SDL_JOYBUTTONUP);
	SDL_FlushEvent(SDL_JOYBUTTONDOWN);
}
