/*******************************************************************
 * 
 *  Author: Peter van Sebille (peter@yipton.net)
 *
 *  (c) Copyright 2016, Peter van Sebille
 *  All Rights Reserved
 *
 *******************************************************************/

#ifndef __SDL_MENU_H
#define __SDL_MENU_H

#include "sdlcore.h"
#include "sdlimenulistener.h"

namespace sdl 
{
	class Menu;

	class MenuItem
	{
	public:
		MenuItem(const std::string& aName, const std::string& aCommand = "") : iName(aName), iCommand(aCommand) {}
		virtual ~MenuItem() {}
		
		const std::string& GetCommand() const	{ return iCommand; }
		const std::string& GetName() const		{ return iName; }
		virtual const Menu* GetSubMenu() const	{ return nullptr; }
		virtual Menu* GetSubMenu() 				{ return nullptr; }
		
		const Menu& GetParentMenu() const ;
		Menu*		GetParentMenu()				{ return iParentMenu; }
		
	private:
		void SetParent(Menu* aParentMenu)		{ iParentMenu = aParentMenu; }
		friend Menu;
		Menu*		iParentMenu = nullptr;
		std::string iName;
		std::string iCommand;
		void*		iPrivateData = nullptr;
	};

	typedef std::vector<std::unique_ptr<MenuItem>> TMenuItemList;

	class Menu : public MenuItem
	{
	public:
		enum TAddPolicy
		{
			EAddToFront,
			EAddToBack
		};
		
		enum TNotifiyPolicy
		{
			ENotifyClients,
			EDontNotifyClients
		};
		
		Menu(const std::string& aName, const std::string& aType = "", int aHighlightedItem = 0) : MenuItem(aName, ""), iType(aType), iPosition(aHighlightedItem) { }
		virtual ~Menu() { }
		
		void AddMenuItemListener(IMenuItemListener& aMenuItemListener, const std::string& aMenuType = "");
		
		void AddMenuItem(MenuItem* aMenuItem, TAddPolicy aAddPolicy = TAddPolicy::EAddToBack, TNotifiyPolicy aNotifiyPolicy = TNotifiyPolicy::ENotifyClients);
		void AddMenuItems(std::vector<std::string>& aList);
		void AddMenu(Menu* aMenu, TAddPolicy aAddPolicy = TAddPolicy::EAddToBack);
		void DeleteMenuItem(const std::string& aCommand, TNotifiyPolicy aNotifiyPolicy = TNotifiyPolicy::ENotifyClients);
		void DeleteAllMenuItems();
		void PruneMenuItems(int aMax, TNotifiyPolicy aNotifiyPolicy = TNotifiyPolicy::ENotifyClients);
		
		void HighlightItem(const std::string& aMenuItem);
		
		bool ScrollUp();
		bool ScrollDown();
		bool ScrollPageUp();
		bool ScrollPageDown();
		
		MenuItem& GetMenuItem(int aPosition);
		const MenuItem& GetMenuItem(int aPosition) const;
		bool HasMenuItem(const std::string& aCommand);
		void Sort();
		void Activate();
		void DeActivate()							{ iIsActive = false; }
		bool IsActive() const						{ return iIsActive; }
		bool IsAtTop() const 						{ return iAtTop; }
		bool IsAtBottom() const 					{ return iAtBottom; }
		MenuItem& GetCurrentMenuItem() 				{ return GetMenuItem(GetCurrentMenuPosition()); }
		const std::string& GetType() const 			{ return iType; }
		int MenuSize() const 						{ return iMenuItems.size(); }
		void SetEnabled(bool aEnabled) 				{ iEnabled = aEnabled; }
		bool IsEnabled() const 						{ return iEnabled; }
		virtual const Menu* GetSubMenu() const override{ return this; }
		virtual Menu* GetSubMenu() override			{ return this; }
		const TMenuItemList& GetMenuItems() const 	{ return iMenuItems; }
		int GetCurrentMenuPosition() const 			{ return iPosition; }
		
		Menu* FindMenu(const std::string& aName);
		void MenuItemCommand(const TCommand& aCommand);
	private:
		void ResetTopAndBottom() 					{ iAtTop = iAtBottom = false;}
		
		unsigned char GetStartCharMenuItem(int aPosition);
		int FindNextMenuItemByStartChar(unsigned char aChar);
		int FindPrevMenuItemByStartChar(unsigned char aChar);
		
		void MenuItemHighlighted();
		void SetHighlightedItem(int aPosition);
		int FindMenuItemPosition(const std::string& aMenuItem);
		void MenuListChanged();
		void KeepInRange();
		
		bool 			iEnabled = true;
		std::string		iType;
		int				iPosition;
		bool 			iAtTop = false;
		bool 			iAtBottom = false;
		int				iMaxItems = -1;
		bool 			iIsActive = false;
		
		TMenuItemList					iMenuItems;
		std::set<IMenuItemListener*>	iMenuItemListeners;
	};

}

#endif