/*******************************************************************
 * 
 *  Author: Peter van Sebille (peter@yipton.net)
 *
 *  (c) Copyright 2016, Peter van Sebille
 *  All Rights Reserved
 *
 *******************************************************************/

#ifndef __SDL_MENU2D_WIDGET_H
#define __SDL_MENU2D_WIDGET_H

#include "sdlcore.h"
#include "sdlcontainer.h"
#include "sdllabelwidget.h"
#include "sdlimagewidget.h"
#include "sdlmenu2dmodel.h"

namespace sdl
{
	class MenuItemWidget : public Container
	{
	public:
		LabelWidget iLabelWidget;
		ImageWidget iImageWidget;
		
		MenuItemWidget(); 
		void SetMenuText(const std::string& aMenuText, bool aIsSubMenu);

	private:
		void SetImagePosition();
		
		virtual void SizeChanged(const TSize& aOldSize) override;
	};
	
	
	class IImageProvider
	{
	public:
		virtual std::string GetFileName(const MenuItem& aItem, const std::string& aLogicalImageDirectory) = 0;
	};
	

	class MenuItemContainer : public Container
	{
	public:
		MenuItemContainer(int aNoOfMenuItems) : iNoOfMenuItems(aNoOfMenuItems) {}
		virtual void Construct(const Menu& aMenu) = 0;
		virtual void VerticalAnimate(int aStep, bool aIsAnimatingTopReached) = 0;
		virtual int GetVerticalAnimateMaxSteps() = 0;
		void Populate(const Menu& aMenu);
		
		
	protected:
		virtual void SetMenuItem(int aIndex, const MenuItem& aMenuItem) = 0;
		
		int		iNoOfMenuItems;
	};
	
	class TextListMenu : public MenuItemContainer
	{
	public:
		enum class TTextStyle
		{
			ELeftAligned = 0,
			ECentered,
			ERound
		};

		class Properties
		{
		public:
			TRect 			iArea;
			int 			iItems = 9;
			int 			iRound;
			std::string 	iTextStyle = "left";
			TFont			iFont;
			int				iFontSizeSelected = 0;
			float			iFontSizeGradient  = 0;
			int				iFontColorGradient = 0;
			bool 			iUseHighlight = false;
			TRoundedRect	iHighlightRoundedRect;
		};

		TextListMenu(const Properties& aProperties) 
			: MenuItemContainer(aProperties.iItems), iTextListMenuProperties(aProperties) {}
		virtual void Construct(const Menu& aMenu) override;
	protected: // from MenuItemContainer
		virtual void VerticalAnimate(int aStep, bool aIsAnimatingTopReached) override;
		virtual int GetVerticalAnimateMaxSteps() override { return ItemHeight(); }
		virtual void SetMenuItem(int aIndex, const MenuItem& aMenuItem) override;
		virtual void CreateCachedTextures(IRenderer& aRenderer) override;
		
	private:
		int ItemHeight() const { return GetHeight() / iTextListMenuProperties.iItems; }
		TRect 	GetItemArea(int aItemIndex, int aDeltaY = 0) const;
		TPoint	GetItemTopLeft(int aItemIndex, int aDeltaY = 0) const { return GetItemArea(aItemIndex, aDeltaY).GetTopLeft(); } 
		int CalcPositiononRoundCurve(int aItemTop, bool aIsMiddleItem) const;
		
		void SetProperties();
		void SetTextStyle(const std::string& aTextStyle);
		MenuItemWidget* ConstructMenuItemWidget(int aDeltaFromMiddle);

		const Properties&				iTextListMenuProperties;
		Widget::THorizontalAlignment	iHorizontalAlignement;
		std::vector<MenuItemWidget*>	iMenuItemWidgets;
		TTextStyle						iTextStyle = TTextStyle::ELeftAligned;
	};

	// not yet fully implemented
	class ImageListMenu : public MenuItemContainer
	{
	public:
		class Properties
		{
		public:
			int 		iItems = 3;
			int 		iSelectedItemX = 100;
			TPoint 		iPreviousItemTopLeftOffset;
			TPoint 		iNextItemTopLeftOffset;
			TSize 		iImageSize = TSize(100, 100);
			float 		iStackedImageRatio = 1;
			int 		iStackWidth = 0;
			std::string	iLogicalImageDirectory;
		};
		ImageListMenu(const Properties& aProperties, IImageProvider& aImageProvider) 
			: MenuItemContainer(aProperties.iItems), iImageListMenuProperties(aProperties), iImageProvider(aImageProvider) {}
		virtual void Construct(const Menu& aMenu) override;
		
	protected: // from MenuItemContainer
		virtual void VerticalAnimate(int aStep, bool aIsAnimatingTopReached) override;
		virtual int GetVerticalAnimateMaxSteps() override { return 20; }
		virtual void SetMenuItem(int aIndex, const MenuItem& aMenuItem) override;
		virtual void Render(IRenderer& aRenderer) override;
		TPoint GetMidPoint(int aItem);
		void VerticalScrollAnimate(int aStep, bool aAnimateUp);
		
	private:		
		const Properties&			iImageListMenuProperties;
		IImageProvider& 			iImageProvider;
		std::vector<ImageWidget*> 	iImages;
	};

	
	class Menu2DWidget : public Container
	{
	public:
		class Properties
		{
		public:
			ImageListMenu::Properties	iImageListMenuProperties;
			TextListMenu::Properties	iTextListMenuProperties;
			TRect						iRect;
			bool 						iEnableImages = false;
			
			bool 		iHorizontalAnimation = true;
			bool		iVerticalAnimation = true;
			int 		iHorizontalAnimationDelay = 10;
			int			iVerticalAnimationDelay = 10;
		};
		
		Menu2DWidget(Menu2DModel& aMenuModel, IImageProvider& aImageProvider) 
			: iMenuModel(aMenuModel), iImageProvider(aImageProvider) { SetDefaultProperties(); }
		~Menu2DWidget() { }
		void Construct();
		void AddImageMenuType(const std::string& aMenuType) { iImageMenuTypes.push_back(aMenuType); }
		void SwitchToMenu(const std::string& aMenuName);

	protected: // from Container
		virtual void Render(IRenderer& aRenderer) override;
		virtual bool HandleCommand(TCommand aCommand) override;

	private:
		void SetDefaultProperties();
		void ApplyProperties();
		
		bool ScrollUp();
		bool ScrollDown();
		bool ScrollPageUp();
		bool ScrollPageDown();
		bool ScrollRight()	{ return ScrollHorizontal(false); }
		bool ScrollLeft()	{ return ScrollHorizontal(true); }
		
		bool ScrollHorizontal(bool aScrollLeft);
		void SwitchToMenu(Menu& aMenu, bool aScrollLeft = true);
		
		MenuItemContainer* ConstructMenuItemWidgets(const Menu& aMenu);
		void PopulateMenu(const Menu& aMenu);

		// menu stuff
		Menu& GetCurrentMenu() { return iMenuModel.CurrentMenu();}
		bool GoBack();
	
		bool IsImageMenu(const Menu& aMenu);

		MenuItemContainer* iCurrentMenuItemContainer = nullptr;
		MenuItemContainer* iNewMenuItemContainer = nullptr;
		
		
		Menu2DModel&				iMenuModel;
		IImageProvider& 			iImageProvider;
		Properties 					iMenu2DProperties;
		std::vector<std::string>	iImageMenuTypes;

		class TVerticalAnimation
		{
		public:
			bool IsAnimating() { return iStep!=0; }
			void StartAnimating(MenuItemContainer* aMenuItemContainer, bool aTopReached, int aDelay);
			void StopAnimating() { iStep=0; iDelay=0; iMaxSteps=0; iMenuItemContainer=nullptr;}
			void Animate();

		private:
			int 				iDelay = 0;
			int  				iStep = 0;
			bool 				iIsAnimatingTopReached = false;
			int					iMaxSteps = 0;
			MenuItemContainer*	iMenuItemContainer = nullptr;
		};
		
		TVerticalAnimation iVerticalAnimation;
		friend TVerticalAnimation;
		void StartVerticalAnimating(bool aTopReached);
		
		void StartHorizontalAnimating(bool aScrollLeft);
		void StopHorizontalAnimating();
		
		class THorizontalAnimation
		{
		public:
			bool IsAnimating() { return iStep != 0; }
			void StartAnimating(Widget& aCurrentListWidget, Widget& aNewListWidget, bool aAnimateLeft, int aDelay);
			void StopAnimating() { iStep = 0; iCurrentListWidget = iNewListWidget = nullptr; }
			void Animate();
		private:
			int 	iStep = 0;
			int 	iDelay;
			float	iDelta;
			TPoint	iStartPositionCurrentWidget;
			TPoint	iStartPositionNewWidget;
			Widget* iCurrentListWidget;
			Widget* iNewListWidget;
			const int iMaxSteps = 20;
		};
		
		THorizontalAnimation iHorizontalAnimation = THorizontalAnimation();
		
		bool IsAnimating() { return iHorizontalAnimation.IsAnimating() || iVerticalAnimation.IsAnimating(); }
	};
}

#endif // __SDL_MENU2D_WIDGET_H

