/*******************************************************************
 * 
 *  Author: Peter van Sebille (peter@yipton.net)
 *
 *  (c) Copyright 2016, Peter van Sebille
 *  All Rights Reserved
 *
 *******************************************************************/

#include "sdlwidget.h"
#include "sdlthemeproperties.h"

using namespace sdl;

void Widget::SetParent(Widget* aParent)
{
	iParent = aParent; 
	
	// This will call PositionChanged() which is important. 
	// By setting a parent our iRect is now contained within the parent's iRect and hence GetScreenPosition() will 
	// return a different TPoint(). This is despite the fact that iRect.iPosition has not changed.
	SetPosition(GetPosition()); 
} 

void Widget::Render(IRenderer& aRenderer)
{
	if (IsEnabled())
	{
		if (HasFocus())
			aRenderer.RenderFillRect(GetScreenRect(), SDLThemeProperties.iDefaultHighlightBackgroundColor);
		
		if (iTextureCache.Size() == 0)
		{
			CreateCachedTextures(aRenderer);
		}
		iTextureCache.Draw(aRenderer, GetScreenPosition());
		
		if (iShowDebugOutline && (SDLThemeProperties.iDebug.iShowOutline || iForcedShowDebugOutline))
			aRenderer.RenderRect(GetScreenRect(), SDLThemeProperties.iDebug.iOutlineColor);
	}
}

void Widget::TTextureCache::Draw(IRenderer& aRenderer, const TPoint& aScreenPosition)
{
	for (auto &cachedTexture : iCachedTextures)
		aRenderer.RenderTexture(*cachedTexture.iTexture.get(), cachedTexture.iPoint + aScreenPosition);
}

void Widget::TTextureCache::SetPosition(int aIndex, const TPoint& aPosition)
{
	if ( (aIndex < 0) || (aIndex >= (int)iCachedTextures.size()))
		return;
	iCachedTextures[aIndex].iPoint = aPosition;
}



void Widget::SetHorizontalAlignment(const std::string& aString)
{
	if (aString.size() == 0)
		return;

	if (aString.compare("left") == 0)
		iHorizontalAlignment = THorizontalAlignment::ELeft;
	else if(aString.compare("right") == 0)
		iHorizontalAlignment = THorizontalAlignment::ERight;
	else if(aString.compare("center") == 0)
		iHorizontalAlignment = THorizontalAlignment::ECenter;
}

void Widget::SetVerticalAlignment(const std::string& aString)
{
	if (aString.size() == 0)
		return;

	if (aString.compare("top") == 0)
		iVerticalAlignment = TVerticalAlignment::ETop;
	else if(aString.compare("bottom") == 0)
		iVerticalAlignment = TVerticalAlignment::EBottom;
	else if(aString.compare("center") == 0)
		iVerticalAlignment = TVerticalAlignment::ECenter;
}


int Widget::GetVerticalAlignedY(int aRenderedHeight, int aAvailableHeight) const
{
	int delta = 0;
	switch (iVerticalAlignment)
	{
		case Widget::TVerticalAlignment::EBottom:
			delta = aAvailableHeight - aRenderedHeight;
			break;
		case Widget::TVerticalAlignment::ECenter:
			delta = (aAvailableHeight - aRenderedHeight) /2;
			break;
		case Widget::TVerticalAlignment::ETop:
			// nothing
		default:
			break;
	}
	return delta;
}


TRect Widget::CalcRenderRect(const TRect& aRect, const TMargin& aMargin) const
{
	TRect widgetRect(GetSize());
	widgetRect.ApplyMargin(aMargin);
	
	TRect rect = aRect;
	
	switch (iHorizontalAlignment)
	{
		case THorizontalAlignment::ELeft:
		default:
		{
			rect.AlignLeft(widgetRect);
			break;
		}
		case THorizontalAlignment::ERight:
		{
			// nothing to do if text is bigger than width
			rect.AlignRight(widgetRect);
			break;
		}
		case THorizontalAlignment::ECenter:
		{
			rect.AlignHCenter(widgetRect);
			break;
		}
	}
	
	switch (iVerticalAlignment)
	{
		case TVerticalAlignment::ETop:
		default:
		{
			rect.AlignTop(widgetRect);
			break;
		}
		case TVerticalAlignment::EBottom:
		{
			rect.AlignBottom(widgetRect);
			break;
		}
		case TVerticalAlignment::ECenter:
		{
			rect.AlignVCenter(widgetRect);
			break;
		}
	}
	
	return rect;
}


void Widget::CacheSetPosition(int aIndex, const TPoint& aPoint)
{
	iTextureCache.SetPosition(aIndex, aPoint);
}


void Widget::CacheAdd(std::shared_ptr<ITexture>& aSharedTexture, const TMargin& aMargin)
{
	ITexture* texture = aSharedTexture.get();
	if (texture)
		iTextureCache.Add(aSharedTexture, CalcRenderRect(texture->GetRect(), aMargin).Position());
}

void Widget::CacheAdd(ITexture& aTexture, const TMargin& aMargin)
{
	iTextureCache.Add(aTexture, CalcRenderRect(aTexture.GetRect(), aMargin).Position());
}

void Widget::CacheAdd(ITexture& aTexture, const TPoint& aPoint)
{
	iTextureCache.Add(aTexture, aPoint);
}
