/*******************************************************************
 * 
 *  Author: Peter van Sebille (peter@yipton.net)
 *
 *  (c) Copyright 2016, Peter van Sebille
 *  All Rights Reserved
 *
 *******************************************************************/

#ifndef __SDL_WIDGET_H
#define __SDL_WIDGET_H

#include "sdlcore.h"
#include "sdlcommands.h"

namespace sdl
{
	class Widget
	{
	public:
		enum class TVerticalAlignment
		{
			ETop = 0,
			EBottom,
			ECenter
		};
		
		enum class THorizontalAlignment
		{
			ELeft = 0,
			ERight,
			ECenter
		};
		
		Widget() { }
		Widget(Widget& aParent) : iParent(&aParent) { }
		virtual ~Widget() {}
		
		// setters
		void SetAlignment(TVerticalAlignment aVerticalAlignment) 	{ iVerticalAlignment = aVerticalAlignment; CacheClear(); }
		void SetAlignment(THorizontalAlignment aHorizontalAlignment){ iHorizontalAlignment = aHorizontalAlignment; CacheClear(); }
		void SetName(const std::string& aName) 						{ iName = aName; }
		
		// setters
		void SetRect(const TRect& aRect) 			{ SetSize(aRect.Size()); SetPosition(aRect.Position()); CacheClear(); }
		void SetSize(int aWidth, int aHeight)		{ SetSize(TSize(aWidth, aHeight)); }
		void SetWidth(int aWidth)					{ SetSize(TSize(aWidth, iRect.Height())); }
		void SetHeight(int aHeight)					{ SetSize(TSize(iRect.Width(), aHeight)); }
		void SetPosition(int aX, int aY)			{ SetPosition(TPoint(aX, aY)); }
		void SetSize(const TSize& aSize)			{ iRect.SetSize(aSize); SizeChanged(aSize); }
		void SetPosition(const TPoint& aPosition)	{ TPoint oldPoint = iRect.Position(); iRect.SetPosition(aPosition); PositionChanged(oldPoint); }
		void SetFocus(bool aFocus)					{ iFocus = aFocus; CacheClear(); FocusChanged(); }
		void SetBackgroundColor(int aColor)			{ iBgColor.OverrideDefaults(aColor); CacheClear();  }
		void SetEnabled(bool aEnabled) 				{ iEnabled = aEnabled; }
		
		// getters
		virtual int		GetMinimumHeight() const	{ return 0; }
		virtual bool 	HasInput() const			{ return false; }
		bool 			IsEnabled() const			{ return iEnabled; }
		const TPoint&	GetPosition() const			{ return iRect.Position(); }
		bool 			HasFocus() const			{ return iFocus; }
		TColor			GetBackgroundColor() const	{ return iBgColor; }
		const TSize&	GetSize() const				{ return iRect.Size(); }
		const TRect&	GetRect() const				{ return iRect; }
		int				GetWidth() const			{ return iRect.Width(); }
		int				GetHeight() const			{ return iRect.Height(); }
		TRect			GetScreenRect() const		{ return TRect(GetScreenPosition(), iRect.Size()); }
		Widget*			GetParent() const			{ return iParent; }
		TPoint			GetScreenPosition() const	{ return (iParent != nullptr) ? iRect.Position() + iParent->GetScreenPosition() + iParent->GetRelativeChildPosition() : iRect.Position(); }
		TPoint 			GetScreenPosition(const TPoint& aPoint) const { return (iParent != nullptr) ? aPoint + iParent->GetScreenPosition() + iParent->GetRelativeChildPosition() : aPoint; }
		
		virtual bool 	HandleCommand(TCommand aCommand){ return false; }

		virtual Widget* FindFocusNext()					{ return nullptr; }
		virtual Widget* FindFocusPrevious()				{ return nullptr; }
		
		virtual void Render(IRenderer& aRenderer);
		virtual void ResetSDLState() 					{ CacheClear(); }
		
		// private data
		void SetPrivateData(int aData) 					{ iDataInt = aData ; }
		void SetPrivateData(const std::string& aData) 	{ iDataString = aData ; }
		void SetPrivateData(void* aData) 				{ iDataPtr = aData ; }
		void GetPrivateData(int& aData) const 			{ aData = iDataInt; }
		void GetPrivateData(std::string& aData) const	{ aData = iDataString; }
		void GetPrivateData(void*& aData) const			{ aData = iDataPtr; }
		
		void SetShowDebugOutline(bool aForced=false) 	{ iShowDebugOutline = true; iForcedShowDebugOutline = aForced; }
	protected:
		std::string iName;

		// Cache control
		void CacheAdd(ITexture& aTexture, const TMargin& aMargin);
		void CacheAdd(std::shared_ptr<ITexture>& aSharedGPUImage, const TMargin& aMargin);
		void CacheAdd(ITexture& aTexture, const TPoint& aPoint);
		void CacheClear() { iTextureCache.Clear(); }
		void CacheSetPosition(int aIndex, const TPoint& aPoint);
		virtual void CreateCachedTextures(IRenderer& aRenderer) { }
		
		// size & position
		TRect CalcRenderRect(const TRect& aRect, const TMargin& aMargin) const;
		virtual TPoint GetRelativeChildPosition() const 			{ return TPoint(0, 0); }
		virtual void SizeChanged(const TSize& aOldSize) 			{ CacheClear(); } 
		virtual void PositionChanged(const TPoint& aOldPosition) 	{ }
		virtual void FocusChanged() 								{ }
		virtual void Layout() 										{ }
		
		// Alignment
		void SetVerticalAlignment(const std::string& aString);
		void SetHorizontalAlignment(const std::string& aString);
		int GetVerticalAlignedY(int aRenderedHeight, int aAvailableHeight) const;
		
		TVerticalAlignment		iVerticalAlignment = TVerticalAlignment::ETop;
		THorizontalAlignment	iHorizontalAlignment = THorizontalAlignment::ELeft;
		
	private:
		friend class Container;
		friend class LayoutManager;

		class TTextureCache
		{
		public:
			TTextureCache(const std::string& aWidgetName) : iWidgetName(aWidgetName) {}
			// aPoint is relative to the the widget's position, so passing TPoint(0, 0) means the texture gets drawn at this->GetPosition()
			void Add(ITexture& aTexture, const TPoint& aPoint) { iCachedTextures.push_back(std::move(TTextureCache::CachedTexture(aTexture, aPoint))); }
			void Add(std::shared_ptr<ITexture>& aSharedTexture, const TPoint& aPoint) { iCachedTextures.push_back(std::move(TTextureCache::CachedTexture(aSharedTexture, aPoint))); }
			void Draw(IRenderer& aRenderer, const TPoint& aScreenPosition);
			void Clear() { iCachedTextures.clear(); }
			int  Size() { return iCachedTextures.size(); }
			void SetPosition(int aIndex, const TPoint& aPosition);
		private:
			class CachedTexture
			{
			public:
				CachedTexture(ITexture& aTexture, const TPoint& aPoint) : iPoint(aPoint), iTexture(&aTexture) { }
				CachedTexture(std::shared_ptr<ITexture>& aSharedTexture, const TPoint& aPoint) : iPoint(aPoint), iTexture(aSharedTexture) { }
				TPoint iPoint;
				std::shared_ptr<ITexture> iTexture;
			private:
				CachedTexture();
			};
		private:
			const std::string& iWidgetName;
			std::vector<CachedTexture>	iCachedTextures;
		};
		TTextureCache	iTextureCache = TTextureCache(iName);
		
		
		void SetParent(Widget* aParent);
		void SetParentOwnsMe(bool aParentOwnsMe){ iParentOwnsMe = aParentOwnsMe; }
		bool GetParentOwnsMe()					{ return iParentOwnsMe; }
		
		bool 	iEnabled 		= true;
		bool	iFocus 			= false;
		bool	iParentOwnsMe 	= false;
		Widget* iParent 		= nullptr;
		TColor	iBgColor 		= 0;
		TRect	iRect			= {0, 0, 0, 0};
		
		int 		iDataInt = 0;
		std::string iDataString;
		void* 		iDataPtr = nullptr;
		bool		iShowDebugOutline = false;
		bool 		iForcedShowDebugOutline = false;
	};
} // namespace sdl
#endif // __SDL_WIDGET_H