/*******************************************************************
 * 
 *  Author: Peter van Sebille (peter@yipton.net)
 *
 *  (c) Copyright 2016, Peter van Sebille
 *  All Rights Reserved
 *
 *******************************************************************/

#ifndef __SIMPLE_TEXT_MODEL_H
#define __SIMPLE_TEXT_MODEL_H

#include <string>
#include <vector>
#include <stdio.h>

namespace simpletext 
{

typedef int TColor;


class TCharFormat
{
public:
	TCharFormat() { Reset(); }
	
	void Reset()
	{
		iFontSize 		= -1;
		iTextColor		= 0x80000000;
		iBold			= false; 
		iItalic			= false;
		iUnderline		= false;
		iStrikethrough	= false;
		iFontName.clear();
	}
	
	std::string 	iFontName;
	int				iFontSize;
	TColor			iTextColor;
	bool			iBold;
	bool 			iItalic;
	bool 			iUnderline;
	bool 			iStrikethrough;
};

class TParaFormat
{
public:
	enum class TAlignment
	{
		ELeft = 0,
		ECenter,
		ERight,
		EJustified
	};
	
	TParaFormat()	{ Reset(); }
	void Reset()	{ iAlign = TAlignment::ELeft; }
	
	TAlignment iAlign;
};

typedef std::vector<std::string> TWords;

class TLineFragment
{
public:
	TLineFragment() 	{ ResetText(); }
	bool IsEmpty()		{ return iWords.size() == 0; }
	void Reset()		{ ResetText(); iCharFormat.Reset(); iDescent = 0; }
	void ResetText()	{ iWords.clear(); iSpaceWidth = 4; }
	
	int 		iSpaceWidth;
	int			iDescent;
	TWords 		iWords;
	TCharFormat	iCharFormat;
};
typedef std::vector<TLineFragment>	TLineFragments;


class TLineInfo
{
public:
	TLineInfo() 			{ ResetParagraph(); }
	bool IsEmpty()			{ return iLineFragments.size()==0; }
	void ResetParagraph()	{ ResetLine(); iParaFormat.Reset(); }
	
	void ResetLine()
	{
		iHeight = 0;
		iWidth = 0;
		iLineDescent = 0;
		iLineFragments.clear();
		iIsLastLineInParagraph = true;
	}
	
	int GetNrWords() const
	{
		int nrWords = 0;
		for (int i=0 ; i< (int) iLineFragments.size() ; i++)
			nrWords += iLineFragments.at(i).iWords.size();
		return nrWords;
	}
	
	int 			iHeight;
	int 			iWidth;
	int 			iLineDescent;
	TParaFormat		iParaFormat;
	TLineFragments	iLineFragments;
	bool			iIsLastLineInParagraph;
};

typedef std::vector<TLineInfo>	TLineInfos;

class IFontProvider
{
public:
	virtual ~IFontProvider() {}
	virtual bool FitText(const TCharFormat& aCharFormat, const std::string& aText, std::string& aReturnText, int aMaxWidth, int& aWidth, int& aHeight, bool aEclipse) = 0;
	virtual int GetDescent(const TCharFormat& aCharFormat) = 0;
	virtual void GetDefaultCharacterFormating(TCharFormat& aCharFormat) = 0;
};


class Model
{
public:
	Model() : iMaxWidth(0) { }
	virtual ~Model() { }

	// Property
	void LimitToOneLine(bool aEclipseLastWord)	{ iLimitToOneLine = true; iEclipseLastWord = aEclipseLastWord; }
	
	// Set the formatted text
	void SetText(const std::string& aText)		{ iText = aText; }

	// Layout the formatted text for the given aMaxWidth
	// aFontProvider provides the functionality to allow simpletext::Model to calculate rendering sizes
	void Layout(int aMaxWidth, IFontProvider& aFontProvider);
	
	// After calling Layout, GetLineInfos returns the datastructure describing how to render the formatted text
	const TLineInfos& GetLineInfos() const		{ return iLineInfos; }
	int GetHeight() const;
private:
	TLineInfos		iLineInfos;
	int 			iMaxWidth;
	std::string		iText;
	bool			iLimitToOneLine = false;
	bool			iEclipseLastWord = false;
};

}

#endif // __SIMPLE_TEXT_MODEL_H