/*******************************************************************
 * 
 *  Author: Peter van Sebille (peter@yipton.net)
 *
 *  (c) Copyright 2016, Peter van Sebille
 *  All Rights Reserved
 *
 *******************************************************************/

#include "simpletextparser.h"

using namespace simpletext;

namespace simpletext 
{
	
class Lex
{
public:
	enum class Token
	{
		EEndOfText = 0,
		ETagStart,
		ETagEnd,
		EString,
		ETagEqual
	};
	Lex(const char* aText) : iText(aText), iCurText(aText), iTokenStart(nullptr), iTokenLen(0), iTagOpen(false), iQuoteOpen(false), iParsingText(false) { }
	
	Token GetToken(int& aStart, int& aLen);
	const char* GetText() { return iText; }
private:
	void AddToText();
	void SingleCharTokenParsed();
	
	const char*	iText;
	const char*	iCurText;
	const char*	iTokenStart;
	int			iTokenLen;
	bool		iTagOpen;
	bool		iQuoteOpen;
	bool		iParsingText;
};
}

void Lex::SingleCharTokenParsed()
{
	iTokenStart = iCurText++;
	iTokenLen = 1;
}

void Lex::AddToText()
{
	if (iTokenStart == nullptr)
	{
		iTokenStart = iCurText;
	}
	++iTokenLen;
	iCurText++;
	iParsingText = true;
}


Lex::Token Lex::GetToken(int& aStart, int& aLen)
{
	Token token = Token::EEndOfText;
	bool gotToken = false;
	
	do
	{
		char c = *iCurText;
		switch (c)
		{
			case 0:
				if (iParsingText)
				{
					token = Token::EString;
					gotToken = true;
				}
				else
				{
					token = Token::EEndOfText;
					SingleCharTokenParsed();
					gotToken = true;
				}
				break;
			
			case '=':
				if (iTagOpen)
				{
					if (iQuoteOpen)
					{
						AddToText();
					}
					else
					{
						if (iParsingText)
						{
							token = Token::EString;
							gotToken = true;
						}
						else
						{
							token = Token::ETagEqual;
							SingleCharTokenParsed();
							gotToken = true;
						}
					}
				}
				else
				{
					AddToText();
				}
				break;
			
			case '"':
				if (iTagOpen)
				{
					if (iQuoteOpen)
					{
						token = Token::EString;
						gotToken = true;
						iQuoteOpen = false;
						++iCurText; // consume quote - but don't add to parsed text
					}
					else
					{
						iQuoteOpen = true;
						++iCurText; // consume quote - but don't add to parsed text
					}
				}
				else
				{
					AddToText();
				}
				break;
				
			case '>':
				if (iTagOpen)
				{
					if (iQuoteOpen)
					{
						AddToText();
					}
					else
					{
						if (iParsingText)
						{
							token = Token::EString;
							gotToken = true;
						}
						else
						{
							iTagOpen = false;
							token = Token::ETagEnd;
							SingleCharTokenParsed();
							gotToken = true;
						}
					}
				}
				else
				{
					AddToText();
				}
				break;
			
			case '<':
				if (iParsingText)
				{
					token = Token::EString;
					gotToken = true;
				}
				else
				{
					iTagOpen = true;
					token = Token::ETagStart;
					SingleCharTokenParsed();
					gotToken = true;
				}
				break;
				
			case ' ':
			case '\t':
			case '\r':
			case '\n':
				if (iTagOpen)
				{
					if (iQuoteOpen)
					{
						AddToText();
					}
					else
					{
						if (iParsingText)
						{
							token = Token::EString;
							gotToken = true;
						}
						else
						{
							++iCurText; // consume  whitespace
						}
					}
				}
				else
				{
					AddToText();
				}
				break;
				
			default:
				AddToText();
				break;
		}
	} while (!gotToken);
	
	aStart = iTokenStart - iText;
	aLen = iTokenLen;
	
	iTokenStart = nullptr;
	iTokenLen = 0;
	iParsingText = false;
	
	return token;
}


void Parser::Parse(const std::string& aText, IParseConsumer& aParseConsumer)
{
	::Lex lex(aText.c_str());
	Parse(lex, aParseConsumer);
}

void Parser::ResetTag()
{
	iTag.Clear();
	iInTag = false;
	iKey.clear();
	iParsedEqual = false;
}

void Parser::Parse(Lex& aLex, IParseConsumer& aParseConsumer)
{
	int start, len;
	Lex::Token token;
	do 
	{
		token = aLex.GetToken(start, len);
		std::string s(aLex.GetText() + start, len);
		switch (token)
		{
			case Lex::Token::EEndOfText:
				if (iInTag)
				{
					aParseConsumer.Warning(start, len, "No closing tag");
				}
				break;
				
			case Lex::Token::ETagStart:
				ResetTag();
				iInTag = true;
				break;
				
			case Lex::Token::EString:
				if (iInTag)
				{
					if (iTag.iName.empty())
					{
						iTag.iName = s;
					}
					else
					{
						if (iKey.empty())
						{
							iKey = s;
						}
						else 
						{
							if (iParsedEqual)
							{
								iTag.iAtributes[iKey] = s;
								iKey.clear();
							}
							else 
							{
								std::string warning("Expected attribute value, ignoring preceding key: ");
								warning += iKey;
								aParseConsumer.Warning(start, len, warning.c_str());
								iKey.clear();
							}
						}
					}
					iParsedEqual = false;
				}
				else
				{
					aParseConsumer.ConsumeText(s);
				}
				break;
				
			case Lex::Token::ETagEnd:
				if (iTag.iName.empty())
				{
					aParseConsumer.Warning(start, len, "Empty tag, ignoring");
				}
				else 
				{
					if (!iKey.empty() || iParsedEqual)
					{
						std::string warning("Expected attribute value, ignoring preceding key: ");
						warning += iKey;
						aParseConsumer.Warning(start, len, warning.c_str());
					}

					aParseConsumer.ConsumeTag(iTag);
				}
				ResetTag();
				break;
				
			case Lex::Token::ETagEqual:
				iParsedEqual = true;
				break;
		}
	} while(token != Lex::Token::EEndOfText);
	
	aParseConsumer.Done();
}

